/*
 * Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 *
 * (c) Copyright 1996 - 2001 Gary Henderson (gary.henderson@ntlworld.com) and
 *                           Jerremy Koot (jkoot@snes9x.com)
 *
 * Super FX C emulator code 
 * (c) Copyright 1997 - 1999 Ivar (ivar@snes9x.com) and
 *                           Gary Henderson.
 * Super FX assembler emulator code (c) Copyright 1998 zsKnight and _Demo_.
 *
 * DSP1 emulator code (c) Copyright 1998 Ivar, _Demo_ and Gary Henderson.
 * C4 asm and some C emulation code (c) Copyright 2000 zsKnight and _Demo_.
 * C4 C code (c) Copyright 2001 Gary Henderson (gary.henderson@ntlworld.com).
 *
 * DOS port code contains the works of other authors. See headers in
 * individual files.
 *
 * Snes9x homepage: http://www.snes9x.com
 *
 * Permission to use, copy, modify and distribute Snes9x in both binary and
 * source form, for non-commercial purposes, is hereby granted without fee,
 * providing that this license information and copyright notice appear with
 * all copies and any derived work.
 *
 * This software is provided 'as-is', without any express or implied
 * warranty. In no event shall the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Snes9x is freeware for PERSONAL USE only. Commercial users should
 * seek permission of the copyright holders first. Commercial use includes
 * charging money for Snes9x or software derived from Snes9x.
 *
 * The copyright holders request that bug fixes and improvements to the code
 * should be forwarded to them so everyone can benefit from the modifications
 * in future versions.
 *
 * Super NES and Super Nintendo Entertainment System are trademarks of
 * Nintendo Co., Limited and its subsidiary companies.
 */
#include "wsnes9x.h"
#include "DirectX.h"
#include "Snes9X/Snes9X.H"
#include "Snes9X/MemMap.H"
#include "Snes9X/cpuexec.h"
#include "Snes9X/display.h"
#include "snes9x/3d.h"
#include "snes9x/cheats.h"
#include "snes9x/netplay.h"
#include "snes9x/apu.h"

//uncomment to find memory leaks, with a line in WinMain
//#include <crtdbg.h>

#include <commctrl.h>

/* For the note-triggered SPC dumping */
int spc_is_dumping = 0;
int spc_is_dumping_temp;
uint8 spc_dump_dsp[0x100];

#include <io.h>
#ifdef RTC_DEBUGGER
#include <time.h>
#endif

extern SNPServer NPServer;

#ifdef USE_OPENGL
OpenGLData OpenGL;
#endif

#include <ctype.h>
#include "fmod.h"

#ifdef RTC_DEBUGGER
int CALLBACK SPC7110rtc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam);
#endif
int CALLBACK SP7PackConfig(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam);
int CALLBACK SoundConf(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam);
int CALLBACK InfoProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam);

#ifdef RTC_DEBUGGER
//Messages for sliders (for some reason they don't get included during the build)
#define TBM_GETPOS			(WM_USER)
#define TBM_GETRANGEMIN		(WM_USER+1)
#define TBM_GETRANGEMAX		(WM_USER+2)
#define TBM_GETTIC			(WM_USER+3)
#define TBM_SETTIC			(WM_USER+4)
#define TBM_SETPOS			(WM_USER+5)
#define TBM_SETRANGE			(WM_USER+6)
#define TBM_SETRANGEMIN		(WM_USER+7)
#define TBM_SETRANGEMAX		(WM_USER+8)
#define TBM_CLEARTICS		(WM_USER+9)
#define TBM_SETSEL			(WM_USER+10)
#define TBM_SETSELSTART		(WM_USER+11)
#define TBM_SETSELEND		(WM_USER+12)
#define TBM_GETPTICS			(WM_USER+14)
#define TBM_GETTICPOS		(WM_USER+15)
#define TBM_GETNUMTICS		(WM_USER+16)
#define TBM_GETSELSTART		(WM_USER+17)
#define TBM_GETSELEND		(WM_USER+18)
#define TBM_CLEARSEL			(WM_USER+19)
#define TBM_SETTICFREQ		(WM_USER+20)
#define TBM_SETPAGESIZE		(WM_USER+21)
#define TBM_GETPAGESIZE		(WM_USER+22)
#define TBM_SETLINESIZE		(WM_USER+23)
#define TBM_GETLINESIZE		(WM_USER+24)
#define TBM_GETTHUMBRECT	(WM_USER+25)
#define TBM_GETCHANNELRECT	(WM_USER+26)
#define TBM_SETTHUMBLENGTH	(WM_USER+27)
#define TBM_GETTHUMBLENGTH	(WM_USER+28)
#endif

#define NOTKNOWN "Unknown Company "

/*****************************************************************************/
/* Global variables                                                          */
/*****************************************************************************/
struct sGUI GUI;

extern FILE *trace_fs;
extern int superscope_turbo;
extern int superscope_pause;
extern SCheatData Cheat;

HINSTANCE g_hInst;

#ifdef DEBUGGER
extern "C" void Trace ();
#endif

static const char *rom_filename = NULL;

CDirectX DirectX;
struct SJoypad Joypad[5] = {
    {
        true,					/* Joypad 1 enabled */
        VK_LEFT, VK_RIGHT, VK_UP, VK_DOWN,	/* Left, Right, Up, Down */
        0, 0,                                   /* Left_Up, Left_Down */
        0, 0,                                   /* Right_Up, Right_Down */
        VK_RETURN, VK_SPACE,                    /* Start, Select */
        'D', 'C',				/* A B */
        'S', 'X',				/* X Y */
        'A', 'Z'				/* L R */
    },
    { 
        false,                                  /* Joypad 2 disabled */
        0, 0, 0, 0,
        0, 0, 0, 0,
        0, 0,
        0, 0, 0, 0, 0, 0
    }, 
    { 
        false,                                  /* Joypad 3 disabled */
        0, 0, 0, 0,
        0, 0, 0, 0,
        0, 0,
        0, 0, 0, 0, 0, 0
    }, 
    { 
        false,                                  /* Joypad 4 disabled */
        0, 0, 0, 0,
        0, 0, 0, 0,
        0, 0,
        0, 0, 0, 0, 0, 0
    }, 
    { 
        false,                                  /* Joypad 5 disabled */
        0, 0, 0, 0,
        0, 0, 0, 0,
        0, 0,
        0, 0, 0, 0, 0, 0
    }
};

struct SSoundRates 
{
    uint32 rate;
    int ident;
} SoundRates[9] = {
    { 8000, ID_SOUND_8000HZ},
    {11025, ID_SOUND_11025HZ},
    {16000, ID_SOUND_16000HZ},
    {22050, ID_SOUND_22050HZ},
    {30000, ID_SOUND_30000HZ},
	{32000, ID_SOUND_32000HZ},
    {35000, ID_SOUND_35000HZ},
    {44100, ID_SOUND_44100HZ},
    {48000, ID_SOUND_48000HZ}
};

// Languages supported by Snes9X: Windows
// 0 - English [Default]
// 1 - Dutch/Nederlands
struct sLanguages Languages[] = {
	{ IDR_MENU_US, 
	  "DirectX failed to initialize!", 
	  "DirectDraw failed to set the selected display mode!", 
	  "DirectSound failed to initialize, no sound will be played.",
	  "These settings won't take effect until you restart the emulator.",
	  "The frame timer failed to initialize, please do NOT select the automatic framerate option or Snes9X will crash!"},
	{ IDR_MENU_NL, 
	  "Er is een fout opgetreden tijdens het initalizeren van DirectX!", 
	  "Er is een fout opgetreden tijdens het verander van scherm modus!", 
	  "Er is een fout opgetreden tijdens het initializeren van DirectSound, er zal geen geluid worden afgespeeld.",
	  "Deze opties worden pas toegepast als de emulator opnieuw is opgestart.",
	  "Er is een fout opgetreden tijdens het initializeren van de frame timer, kies NIET de automatische framerate optie want dan zal Snes9X crashen!"}
#ifdef SEE_JAMES
,
{ IDR_MENU_DE,
"DirectX kann nicht initialisiert werden"
"DirectX kann die Anzeige nicht ndern"
"Starten Sie Snes9x neu um die Eigenschaften zu bernehmen"
"Frame Timer konnte nicht initialisiert werden, NICHT `automatic framerate' option verwenden oder Snes9x strzt ab!"}};
#else
};
#endif

extern "C" {
    bool __cdecl _OpenROM( char *, struct SSettings *);
    void __cdecl _JoypadConfig( struct SJoypad *);
    void __cdecl _CheatCodeEntry( struct SCheatData *);
    void __cdecl _InitDLL( HANDLE hWnd);
    void __cdecl _DirectXConfig( LPDIRECTDRAW plpDD, struct SSettings *Settings, struct sGUI *GUI, bool *showFPS);
    int __cdecl _GUIVersion ();
    void __cdecl _SearchForCheats (struct SCheatData *);
    void __cdecl _SettingsDialog (struct SSettings *, struct sGUI *);
    void __cdecl _AboutDialog (const char *);
    bool __cdecl _NetPlayConnect (char *, int *);
    bool __cdecl _NetPlayOptions (struct SSettings *, struct SNetPlay *,
                                  struct SNPServer *);
    bool __cdecl _SoundOptions (struct SSettings *);
    void __cdecl _ProgressDialog (const char *message, int progress);
    HWND __cdecl _MainWindow ();
};

/*****************************************************************************/
/* WinProc                                                                   */
/*****************************************************************************/
bool8 SetupSound( long rate, bool8 sixteen_bit, bool8 stereo);
void RestoreGUIDisplay ();
void RestoreSNESDisplay ();
void FreezeUnfreeze (int slot, bool8 freeze);
void CheckDirectoryIsWritable (const char *filename);
static void CheckMenuStates ();
static void ResetFrameTimer ();
bool8 LoadROM (const char *filename);
static void EnableServer (bool8 enable);
static void WinDeleteRegistryEntries ();
static void WinSetDefaultValues ();
static void WinDeleteRecentGamesList ();

#ifdef USE_DIRECTX3D
void D3DMove (HWND, LPARAM);
void D3DSizeChange (HWND, WPARAM);
#endif

#include "snes9x/ppu.h"
#include "snes9x/snapshot.h"
extern "C" const char *S9xGetFilename (const char *);
const char *S9xGetFilenameInc (const char *);
void S9xSetRecentGames ();
void S9xAddToRecentGames (const char *filename);

void S9xMouseOn ()
{
    if (IPPU.Controller == SNES_MOUSE || IPPU.Controller == SNES_MOUSE_SWAPPED)
    {
        SetCursor (NULL);
    }
    else
    if (IPPU.Controller != SNES_SUPERSCOPE)
    {
        SetCursor (GUI.Arrow);
        GUI.CursorTimer = 60;
    }
    else
        SetCursor (GUI.GunSight);
}
            
void S9xRestoreWindowTitle ()
{
    char buf [100];
    sprintf (buf, "Snes9X v%s for Windows 95, NT 4.0, and higher", VERSION);
    SetWindowText (GUI.hWnd, buf);
}

void SwitchToGDI()
{
    if (!VOODOO_MODE && !OPENGL_MODE)
    {
        IPPU.ColorsChanged = true;
        DirectX.lpDD->FlipToGDISurface();
        GUI.FlipCounter = 0;
        DirectX.lpDDSPrimary2->SetPalette (NULL);
    }
}

static void S9xClearSurface (LPDIRECTDRAWSURFACE2 lpDDSurface)
{
    DDBLTFX fx;

    memset (&fx, 0, sizeof (fx));
    fx.dwSize = sizeof (fx);

    while (lpDDSurface->Blt (NULL, DirectX.lpDDSPrimary2, NULL, DDBLT_WAIT, NULL) == DDERR_SURFACELOST)
        lpDDSurface->Restore ();
}

void UpdateBackBuffer()
{
    GUI.ScreenCleared = true;
    
    if (!VOODOO_MODE && !OPENGL_MODE && GUI.FullScreen)
    {
        SwitchToGDI();

        DDBLTFX fx;

        memset (&fx, 0, sizeof (fx));
        fx.dwSize = sizeof (fx);

        while (DirectX.lpDDSPrimary2->Blt (NULL, NULL, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &fx) == DDERR_SURFACELOST)
            DirectX.lpDDSPrimary2->Restore ();
        
        if (GetMenu (GUI.hWnd) != NULL)
            DrawMenuBar (GUI.hWnd);

        GUI.FlipCounter = 0;
        DDSCAPS caps;
        caps.dwCaps = DDSCAPS_BACKBUFFER;

        LPDIRECTDRAWSURFACE2 pDDSurface;

        if (DirectX.lpDDSPrimary2->GetAttachedSurface (&caps, &pDDSurface) == DD_OK &&
            pDDSurface != NULL)
        {
            S9xClearSurface (pDDSurface);
            DirectX.lpDDSPrimary2->Flip (NULL, DDFLIP_WAIT);
            while (DirectX.lpDDSPrimary2->GetFlipStatus (DDGFS_ISFLIPDONE) != DD_OK)
                Sleep (0);
            S9xClearSurface (pDDSurface);
        }
    }
    else
    { 
        if (GetMenu( GUI.hWnd) != NULL)
            DrawMenuBar (GUI.hWnd);
    }
}

void ToggleFullScreen ()
{
    S9xSetPause (PAUSE_TOGGLE_FULL_SCREEN);

#ifdef USE_GLIDE
    if (VOODOO_MODE)
    {
//        S9xGlideEnable (FALSE);
//        GUI.Scale = 0;
//        MoveWindow (GUI.hWnd, GUI.window_size.left,
//                    GUI.window_size.top, 
//                    GUI.window_size.right - GUI.window_size.left,
//                    GUI.window_size.bottom - GUI.window_size.top, TRUE);
    }
    else
    if (OPENGL_MODE)
    {
        if (!GUI.FullScreen)
        {
            DEVMODE dmScreenSettings;

            memset (&dmScreenSettings, 0, sizeof(dmScreenSettings));
            dmScreenSettings.dmSize = sizeof(dmScreenSettings);
            dmScreenSettings.dmPelsWidth = GUI.Width;
            dmScreenSettings.dmPelsHeight = GUI.Height;
            dmScreenSettings.dmBitsPerPel = 16; //bits;
            dmScreenSettings.dmFields = DM_BITSPERPEL | DM_PELSWIDTH |
                                        DM_PELSHEIGHT;
            if (ChangeDisplaySettings (&dmScreenSettings, CDS_FULLSCREEN) == DISP_CHANGE_SUCCESSFUL)
            {
                GUI.FullScreen = TRUE;
                SetWindowLong (GUI.hWnd, GWL_STYLE, WS_POPUP | WS_VISIBLE);
                SetWindowPos (GUI.hWnd, HWND_TOP, 0, 0, GUI.Width,
                              GUI.Height,
                              SWP_DRAWFRAME|SWP_FRAMECHANGED);
            }
        }
        else
        {
            SetWindowLong (GUI.hWnd, GWL_STYLE, WS_POPUPWINDOW|WS_CAPTION|
                           WS_THICKFRAME|WS_VISIBLE|WS_MINIMIZEBOX|WS_MAXIMIZEBOX);
            SetWindowPos (GUI.hWnd, HWND_TOP, 
                          GUI.window_size.left,
                          GUI.window_size.top,
                          GUI.window_size.right - GUI.window_size.left,
                          GUI.window_size.bottom - GUI.window_size.top,
                          SWP_DRAWFRAME|SWP_FRAMECHANGED);
            ChangeDisplaySettings (NULL, 0);
            GUI.FullScreen = FALSE;
        }
    }
    else
    {
#endif
        if (!VOODOO_MODE && !GUI.FullScreen)
            GetWindowRect (GUI.hWnd, &GUI.window_size);
        
        if (!DirectX.SetDisplayMode (GUI.Width, GUI.Height, GUI.Depth, GUI.FullScreen, GUI.DoubleBuffered))
        {
            MessageBox( GUI.hWnd, Languages[ GUI.Language].errModeDD, "Snes9X - DirectDraw(2)", MB_OK | MB_ICONSTOP);
            S9xClearPause (PAUSE_TOGGLE_FULL_SCREEN);
            return;
        }
        GUI.FullScreen = !GUI.FullScreen;
        if (!GUI.FullScreen)
        {
            SwitchToGDI();
            MoveWindow (GUI.hWnd, GUI.window_size.left,
                        GUI.window_size.top,
                        GUI.window_size.right - GUI.window_size.left,
                        GUI.window_size.bottom - GUI.window_size.top, TRUE);
        }
#ifdef USE_GLIDE
    }
#endif
    S9xGraphicsDeinit();
    S9xSetWinPixelFormat ();
    S9xGraphicsInit();

    IPPU.RenderThisFrame = true;
    UpdateBackBuffer();

    S9xClearPause (PAUSE_TOGGLE_FULL_SCREEN);
}

void S9xDisplayStateChange (const char *str, bool8 on)
{
    static char string [100];

    sprintf (string, "%s %s", str, on ? "on" : "off");
    S9xSetInfoString (string);
}

LRESULT CALLBACK WinProc(
	HWND hWnd,
	UINT uMsg,
	WPARAM wParam,
	LPARAM lParam)
{
    int i;
    bool showFPS;
    static char InfoString [100];
    char hostname [100];

    switch (uMsg)
    {
	case WM_CREATE:
		g_hInst = ((LPCREATESTRUCT)lParam)->hInstance;
		return 0;
	case WM_KEYDOWN:
            switch (wParam)
            {
                case VK_TAB:
					//until I decide how to really handle this
					if(Settings.SPC7110RTC)
						break;
                    if (GUI.TurboModeToggle)
                    {
                        Settings.TurboMode ^= TRUE;
                        if (Settings.TurboMode)
                            S9xMessage (S9X_INFO, S9X_TURBO_MODE,
                                        "Turbo Mode Activated");
                        else
                            S9xMessage (S9X_INFO, S9X_TURBO_MODE,
                                        "Turbo Mode Off");
                    }
                    else
                    {
                        Settings.TurboMode = TRUE;
                        S9xMessage (S9X_INFO, S9X_TURBO_MODE, "Turbo Mode");
                    }
                        
                    break;
		case VK_ESCAPE:
                    if (GetMenu (GUI.hWnd) == NULL)
                        SetMenu (GUI.hWnd, GUI.hMenu);
                    else
                        SetMenu (GUI.hWnd, NULL);

                    UpdateBackBuffer();
                    break;

                case VK_SCROLL:
                    break;
		case '0':
                    Settings.DisableHDMA = !Settings.DisableHDMA;
		    S9xDisplayStateChange ("HDMA emulation", !Settings.DisableHDMA);
                    break;
		case '1':
                    PPU.BG_Forced ^= 1;
		    S9xDisplayStateChange ("BG#0", !(PPU.BG_Forced & 1));
                    break;
		case '2':
                    PPU.BG_Forced ^= 2;
		    S9xDisplayStateChange ("BG#1", !(PPU.BG_Forced & 2));
                    break;
		case '3':
                    PPU.BG_Forced ^= 4;
		    S9xDisplayStateChange ("BG#2", !(PPU.BG_Forced & 4));
                    break;
		case '4':
                    PPU.BG_Forced ^= 8;
		    S9xDisplayStateChange ("BG#3", !(PPU.BG_Forced & 8));
                    break;
		case '5':
                    PPU.BG_Forced ^= 16;
		    S9xDisplayStateChange ("Sprites", !(PPU.BG_Forced & 16));
                    break;
		case '6':
#ifdef USE_OPENGL
                    if (Settings.OpenGLEnable && 
                        (GetKeyState (VK_SHIFT) & 0x80))
                    {
                        OpenGL.draw_cube ^= TRUE;
                    }
                    else
#endif
                    {
                        Settings.SwapJoypads = !Settings.SwapJoypads;
                        S9xDisplayStateChange ("Joypad swapping", Settings.SwapJoypads);
                    }
                    break;
                case '7':
                {
		    static char *controllers [] = {
			"Multiplayer 5 on #0", "Joypad on #0", "Mouse on #1",
			"Mouse on #0", "Superscope on #1"
		    };
		    S9xNextController ();
		    S9xSetInfoString (controllers [IPPU.Controller]);
                    if (IPPU.Controller == SNES_SUPERSCOPE)
                    {
                        SetCursor (GUI.GunSight);
                    }
                    else
                    {
                        SetCursor (GUI.Arrow);
                        GUI.CursorTimer = 60;
                    }

                    if (IPPU.Controller == SNES_MOUSE || 
                        IPPU.Controller == SNES_MOUSE_SWAPPED)
                    {
                        POINT middle;
                        RECT size;

                        GetClientRect (GUI.hWnd, &size);
                        middle.x = (size.right - size.left) >> 1;
                        middle.y = (size.bottom - size.top) >> 1;
                        ClientToScreen (GUI.hWnd, &middle);
                        SetCursorPos (middle.x, middle.y);
                    }
                    break;
                }
		case '8':
                    Settings.BGLayering = !Settings.BGLayering;
		    S9xDisplayStateChange ("Background layering hack", 
					   Settings.BGLayering);
                    break;
		case '9':
                    if (GetKeyState (VK_SHIFT) & 0x80)
                    {
			Settings.Mode7Interpolate ^= TRUE;
			S9xDisplayStateChange ("Mode 7 Interpolation", 
					       Settings.Mode7Interpolate);
                    }
                    else
                    if (Settings.SixteenBit)
                    {	
                        Settings.Transparency = !Settings.Transparency;
			S9xDisplayStateChange ("Transparency effects", 
					       Settings.Transparency);
                    }
                    break;

		case VK_BACK:
                    Settings.DisableGraphicWindows = !Settings.DisableGraphicWindows;
		    S9xDisplayStateChange ("Graphic clip windows",
					   !Settings.DisableGraphicWindows);
                    break;

		case VK_PAUSE:
                    Settings.Paused = Settings.Paused ^ true;
		    S9xDisplayStateChange ("Pause", Settings.Paused);
                    break;

		case VK_F1:
		case VK_F2:
		case VK_F3:
		case VK_F4:
		case VK_F5:
		case VK_F6:
		case VK_F7:
		case VK_F8:
		case VK_F9:
                case VK_F10:
                    FreezeUnfreeze (wParam - VK_F1,
                                    (GetKeyState (VK_SHIFT) & 0x80) != 0);
                    break;
            }
            break;

	case WM_KEYUP:
            switch( wParam)
            {
                case VK_TAB:
                    if (!GUI.TurboModeToggle)
                        Settings.TurboMode = FALSE;
                    break;
            }
            break;
        case WM_CHAR:
            switch (wParam)
            {
                case '`':
                case '#':
                case '~':
                    superscope_turbo ^= 1;
                    break;
                case '/':
                    superscope_pause ^= 1;
                    break;
                case '+':
                case '=':
                    if (GetAsyncKeyState (VK_SHIFT))
                    {
			// Increase emulated frame time by 1ms
			Settings.FrameTime += 1;
                        ResetFrameTimer ();
			sprintf (InfoString, "Emulated frame time: %dms",
				 Settings.FrameTime / 1);
			S9xSetInfoString (InfoString);
		    }
		    else
		    {
			if (Settings.SkipFrames == AUTO_FRAMERATE)
			    Settings.SkipFrames = 1;
			else
			if (Settings.SkipFrames < 10)
			    Settings.SkipFrames++;

			if (Settings.SkipFrames == AUTO_FRAMERATE)
			    S9xSetInfoString ("Auto frame skip");
			else
			{
			    sprintf (InfoString, "Frame skip: %d",
				     Settings.SkipFrames - 1);
			    S9xSetInfoString (InfoString);
			}
                    }
                    break;
                case '-':
                case '_':
                    if (GetAsyncKeyState (VK_SHIFT))
		    {
			// Decrease emulated frame time by 1ms
			if (Settings.FrameTime >= 1)
			    Settings.FrameTime -= 1;

                        ResetFrameTimer ();
			sprintf (InfoString, "Emulated frame time: %dms",
				 Settings.FrameTime / 1);
			S9xSetInfoString (InfoString);
		    }
		    else
		    {
			if (Settings.SkipFrames <= 1)
			    Settings.SkipFrames = AUTO_FRAMERATE;
			else
			    if (Settings.SkipFrames != AUTO_FRAMERATE)
				Settings.SkipFrames--;

			if (Settings.SkipFrames == AUTO_FRAMERATE)
			    S9xSetInfoString ("Auto frame skip");
			else
			{
			    sprintf (InfoString, "Frame skip: %d",
				     Settings.SkipFrames - 1);
			    S9xSetInfoString (InfoString);
			}
		    }
                    break;
            }
            break;
	case WM_COMMAND:
            switch (wParam & 0xffff)
            {
		case ID_OPTIONS_DISPLAY:
                {
                    int old_scale = GUI.NextScale;
                    bool old_stretch = GUI.Stretch;

                    RestoreGUIDisplay ();
                    showFPS = Settings.DisplayFrameRate ? true : false;
                    if (!VOODOO_MODE && !GUI.FullScreen)
                        GetWindowRect (GUI.hWnd, &GUI.window_size);
                    _DirectXConfig (DirectX.lpDD, &Settings, &GUI, &showFPS);

                    Settings.DisplayFrameRate = showFPS;
                    SwitchToGDI();
                    if (GUI.NextScale != old_scale)
                    {
                        if ((GUI.Scale < 7 && GUI.NextScale < 7) ||
                            (GUI.Scale > 7 && GUI.NextScale > 7))
                            GUI.Scale = GUI.NextScale;
                        else
                        {
                            MessageBox (GUI.hWnd, "\
You will need to restart Snes9x before the output image\n\
processing option change will take effect.",
                                        "Snes9x: Information",
                                        MB_OK | MB_ICONINFORMATION);
                        }
                    }
                    RestoreSNESDisplay ();
                    if (!GUI.FullScreen)
                    {
                        MoveWindow (GUI.hWnd, GUI.window_size.left,
                                    GUI.window_size.top, 
                                    GUI.window_size.right - GUI.window_size.left,
                                    GUI.window_size.bottom - GUI.window_size.top, TRUE);
                    }
                    S9xGraphicsDeinit();
                    S9xSetWinPixelFormat ();
                    S9xGraphicsInit();

                    IPPU.RenderThisFrame = false;

                    if (old_stretch != GUI.Stretch || old_scale != GUI.Scale)
                    {
                        RECT rect;
                        GetClientRect (GUI.hWnd, &rect);
                        InvalidateRect (GUI.hWnd, &rect, true);
                    }
                    break;
                }
		case ID_OPTIONS_JOYPAD:
                    RestoreGUIDisplay ();
                    _JoypadConfig( Joypad);
                    RestoreSNESDisplay ();
                    break;

		case ID_FILE_OPEN:
                {
                    char filename [_MAX_PATH];
				
#ifdef NETPLAY_SUPPORT
                    if (Settings.NetPlay && !Settings.NetPlayServer)
                    {
                        S9xMessage (S9X_INFO, S9X_NETPLAY_NOT_SERVER,
                                    "Disconnect from the NetPlay server first.");
                        break;
                    }
#endif
                    RestoreGUIDisplay ();

                    if (_OpenROM (filename, &Settings))
                    {
                        if (!Settings.StopEmulation)
                        {
                            Memory.SaveSRAM (S9xGetFilename (".srm"));
                            S9xSaveCheatFile (S9xGetFilename (".cht"));
                        }
                        Settings.StopEmulation = !LoadROM (filename);
                        if (!Settings.StopEmulation) 
                        {
                            Memory.LoadSRAM (S9xGetFilename (".srm"));
                            S9xLoadCheatFile (S9xGetFilename (".cht"));
                            S9xAddToRecentGames (filename);
                            CheckDirectoryIsWritable (S9xGetFilename (".---"));
                            CheckMenuStates ();
#ifdef NETPLAY_SUPPORT
                            if (NPServer.SendROMImageOnConnect)
                                S9xNPServerQueueSendingROMImage ();
                            else
                                S9xNPServerQueueSendingLoadROMRequest (Memory.ROMName);
#endif
                        }

                        if (IPPU.Controller == SNES_SUPERSCOPE)
                            SetCursor (GUI.GunSight);
                        else
                        {
                            SetCursor (GUI.Arrow);
                            GUI.CursorTimer = 60;
                        }
                        Settings.Paused = false;
                    }

                    RestoreSNESDisplay ();
                    GUI.ScreenCleared = true;
                }
                break;

		case ID_FILE_EXIT:
                    S9xSetPause (PAUSE_EXIT);
#ifdef USE_GLIDE
                    S9xGlideEnable (FALSE);
#endif
                    PostMessage (hWnd, WM_DESTROY, 0, 0);
                    break;

		case ID_WINDOW_HIDEMENUBAR:
                    if( GetMenu( GUI.hWnd) == NULL)
                        SetMenu( GUI.hWnd, GUI.hMenu);
                    else
                        SetMenu( GUI.hWnd, NULL);
                    GUI.ScreenCleared = true;
                    break;
#ifdef SEE_JAMES
		case ID_LANGUAGE_GERMAN:
					GUI.Language = 2;
                    SetMenu( GUI.hWnd, LoadMenu( GUI.hInstance, MAKEINTRESOURCE( Languages[ GUI.Language].idMenu)));
                    DestroyMenu( GUI.hMenu);
                    GUI.hMenu = GetMenu( GUI.hWnd);
                    break;
#endif
		case ID_LANGUAGE_ENGLISH:
                    GUI.Language = 0;

                    SetMenu( GUI.hWnd, LoadMenu( GUI.hInstance, MAKEINTRESOURCE( Languages[ GUI.Language].idMenu)));
                    DestroyMenu( GUI.hMenu);
                    GUI.hMenu = GetMenu( GUI.hWnd);
                    break;
		case ID_LANGUAGE_NEDERLANDS:
                    GUI.Language = 1;

                    SetMenu( GUI.hWnd, LoadMenu( GUI.hInstance, MAKEINTRESOURCE( Languages[ GUI.Language].idMenu)));
                    DestroyMenu( GUI.hMenu);
                    GUI.hMenu = GetMenu( GUI.hWnd);
                    break;

		case ID_NETPLAY_SERVER:
                    S9xRestoreWindowTitle ();
                    EnableServer (!Settings.NetPlayServer);
                    break;
                case ID_NETPLAY_CONNECT:
                    RestoreGUIDisplay ();
                    if (_NetPlayConnect (hostname, &Settings.Port))
                    {
                        S9xSetPause (PAUSE_NETPLAY_CONNECT);
                        if (!S9xNPConnectToServer (hostname, Settings.Port,
                                                   Memory.ROMName))
                        {
                            S9xClearPause (PAUSE_NETPLAY_CONNECT);
                        }
                    }
                    RestoreSNESDisplay ();
                    break;
                case ID_NETPLAY_DISCONNECT:
                    if (Settings.NetPlay)
                    {
                        Settings.NetPlay = FALSE;
                        S9xNPDisconnect ();
                    }
                    if (Settings.NetPlayServer)
                    {
                        Settings.NetPlayServer = FALSE;
                        S9xNPStopServer ();
                    }
                    break;
                case ID_NETPLAY_OPTIONS:
                {
                    bool8 old_netplay_server = Settings.NetPlayServer;
                    RestoreGUIDisplay ();
                    if (_NetPlayOptions (&Settings, &NetPlay, &NPServer))
                    {
                        if (old_netplay_server != Settings.NetPlayServer)
                        {
                            Settings.NetPlayServer = old_netplay_server;
                            S9xRestoreWindowTitle ();
                            EnableServer (!Settings.NetPlayServer);
                        }
                    }
                    RestoreSNESDisplay ();
                    break;
                }
                case ID_NETPLAY_SYNC:
                    S9xNPServerQueueSyncAll ();
                    break;
                case ID_NETPLAY_ROM:
                    if (NPServer.SyncByReset)
                    {
                        if (MessageBox (GUI.hWnd, "\
Your game will be reset after the ROM has been sent due to\n\
your 'Sync Using Reset Game' setting.\n\n", "Snes9X: Warning",
                                        MB_OKCANCEL | MB_ICONWARNING) == IDCANCEL)
                            break;
                    }
                    S9xNPServerQueueSendingROMImage ();
                    break;
                case ID_NETPLAY_SEND_ROM_ON_CONNECT:
                    NPServer.SendROMImageOnConnect ^= TRUE;
                    break;
                case ID_NETPLAY_SYNC_BY_RESET:
                    NPServer.SyncByReset ^= TRUE;
                    break;
                case ID_SOUND_8000HZ:
		case ID_SOUND_11025HZ:
		case ID_SOUND_16000HZ:
		case ID_SOUND_22050HZ:
		case ID_SOUND_30000HZ:
		case ID_SOUND_35000HZ:
		case ID_SOUND_44100HZ:
		case ID_SOUND_48000HZ:
		case ID_SOUND_32000HZ:
                    for( i = 0; i < COUNT(SoundRates); i ++)
			if (SoundRates[i].ident == (int) wParam)
			{
                            Settings.SoundPlaybackRate = SoundRates [i].rate;
                            if (!SetupSound (Settings.SoundPlaybackRate, 
                                             Settings.SixteenBitSound, 
                                             Settings.Stereo))
                            {	MessageBox( GUI.hWnd, Languages[ GUI.Language].errInitDS, "Snes9X - DirectSound", MB_OK | MB_ICONINFORMATION);	}
                            break;
			}
                    break;

		case ID_SOUND_NOSOUND:
                    Settings.SoundPlaybackRate = 0;
                    SetupSound( 0, FALSE, FALSE);
                    break;
                case ID_SOUND_25MS:
                    Settings.SoundBufferSize = 1;
                    SetupSound (Settings.SoundPlaybackRate, 
                                Settings.SixteenBitSound, Settings.Stereo);
                    break;
                case ID_SOUND_50MS:
                    Settings.SoundBufferSize = 2;
                    SetupSound (Settings.SoundPlaybackRate, 
                                Settings.SixteenBitSound, Settings.Stereo);
                    break;
                case ID_SOUND_100MS:
                    Settings.SoundBufferSize = 4;
                    SetupSound (Settings.SoundPlaybackRate, 
                                Settings.SixteenBitSound, Settings.Stereo);
                    break;
                case ID_SOUND_200MS:
                    Settings.SoundBufferSize = 8;
                    SetupSound (Settings.SoundPlaybackRate, 
                                Settings.SixteenBitSound, Settings.Stereo);
                    break;
                case ID_SOUND_500MS:
                    Settings.SoundBufferSize = 16;
                    SetupSound (Settings.SoundPlaybackRate, 
                                Settings.SixteenBitSound, Settings.Stereo);
                    break;
                case ID_SOUND_1S:
                    Settings.SoundBufferSize = 32;
                    SetupSound (Settings.SoundPlaybackRate, 
                                Settings.SixteenBitSound, Settings.Stereo);
                    break;
                case ID_SOUND_2S:
                    Settings.SoundBufferSize = 64;
                    SetupSound (Settings.SoundPlaybackRate, 
                                Settings.SixteenBitSound, Settings.Stereo);
                    break;
                
                case ID_SOUND_STEREO:
                    Settings.Stereo = !Settings.Stereo;
                    SetupSound (Settings.SoundPlaybackRate, 
                                Settings.SixteenBitSound, Settings.Stereo);
                    break;
                case ID_SOUND_REVERSE_STEREO:
                    Settings.ReverseStereo = !Settings.ReverseStereo;
                    break;
                case ID_SOUND_16BIT:
                    Settings.SixteenBitSound = !Settings.SixteenBitSound;
                    SetupSound (Settings.SoundPlaybackRate, 
                                Settings.SixteenBitSound, Settings.Stereo);
                    break;
                case ID_SOUND_INTERPOLATED:
                    Settings.InterpolatedSound = !Settings.InterpolatedSound;
		    S9xDisplayStateChange ("Interpolated sound", Settings.InterpolatedSound);
                    break;
                case ID_SOUND_SYNC:
                    Settings.SoundSync = !Settings.SoundSync;
		    S9xDisplayStateChange ("Sync sound", Settings.SoundSync);
                    break;
                case ID_SOUND_OPTIONS:
                {
                    struct SSettings orig = Settings;
                    RestoreGUIDisplay ();
                    //if (_SoundOptions (&Settings))
					if(1<=DialogBoxParam(g_hInst,MAKEINTRESOURCE(IDD_SOUND_OPTS),hWnd,SoundConf, (LPARAM)&Settings))
                    {
                        if (orig.NextAPUEnabled != Settings.NextAPUEnabled)
                        {
                            if (!Settings.NextAPUEnabled)
                            {
                                if (MessageBox (GUI.hWnd, "\
Disabling the sound CPU emulation will help to improve\n\
emulation speed but you will not hear any sound effects\n\
or music. If you later want to re-enable the sound CPU\n\
emulation you will need to reset your game before it will\n\
take effect.\n\n\
Are you sure this is what you want?",
                                                "Snes9X - Sound CPU Question",
                                                MB_YESNO | MB_ICONQUESTION) == IDNO)
                                {
                                    Settings.NextAPUEnabled = orig.NextAPUEnabled;
                                }
                                else
                                {
                                    Settings.APUEnabled = FALSE;
                                    SetupSound (0, FALSE, FALSE);
                                }
                            }
                            else
                            {
                                if (!Settings.StopEmulation)
                                {
                                    MessageBox (GUI.hWnd, "\
You will need to reset your game or load another one\n\
before enabling the sound CPU will take effect.", "Snes9X - Sound CPU Information",
                                                MB_OK | MB_ICONINFORMATION);
                                }
                                else
                                    Settings.APUEnabled = Settings.NextAPUEnabled;
                            }
                        }
                        else
                        if (orig.SixteenBitSound != Settings.SixteenBitSound ||
                            orig.Stereo != Settings.Stereo ||
                            orig.SoundPlaybackRate != Settings.SoundPlaybackRate ||
                            orig.SoundBufferSize != Settings.SoundBufferSize ||
                            orig.SoundDriver != Settings.SoundDriver ||
                            orig.SoundMixInterval != Settings.SoundMixInterval)
                        {
                            if (Settings.SoundDriver != WIN_SNES9X_DIRECT_SOUND_DRIVER)
                                Settings.SoundSync = FALSE;

                            SetupSound (Settings.SoundPlaybackRate, 
                                        Settings.SixteenBitSound, 
                                        Settings.Stereo);
                        }
                    }
                    RestoreSNESDisplay ();
                    break;
                }
#ifdef RTC_DEBUGGER
								case IDM_7110_RTC:
				{
					struct SPC7110RTC origrtc = s7r.rtc;
                    RestoreGUIDisplay ();
					if(1<=DialogBoxParam(g_hInst,MAKEINTRESOURCE(IDD_7110_RTC),hWnd,SPC7110rtc, (LPARAM)&origrtc))
					{
						rtc_f9.reg[0x00]=origrtc.reg[0x00];
						rtc_f9.reg[0x01]=origrtc.reg[0x01];
						rtc_f9.reg[0x02]=origrtc.reg[0x02];
						rtc_f9.reg[0x03]=origrtc.reg[0x03];
						rtc_f9.reg[0x04]=origrtc.reg[0x04];
						rtc_f9.reg[0x05]=origrtc.reg[0x05];
						rtc_f9.reg[0x06]=origrtc.reg[0x06];
						rtc_f9.reg[0x07]=origrtc.reg[0x07];
						rtc_f9.reg[0x08]=origrtc.reg[0x08];
						rtc_f9.reg[0x09]=origrtc.reg[0x09];
						rtc_f9.reg[0x0A]=origrtc.reg[0x0A];
						rtc_f9.reg[0x0B]=origrtc.reg[0x0B];
						rtc_f9.reg[0x0C]=origrtc.reg[0x0C];
						
						rtc_f9.reg[0x0D]=origrtc.reg[0x0D];
						
						rtc_f9.reg[0x0E]=origrtc.reg[0x0E];
						rtc_f9.reg[0x0F]=origrtc.reg[0x0F];
						rtc_f9.last_used=time(NULL);
						
					}
					RestoreSNESDisplay ();
                    break;
				}
#endif
                case ID_WINDOW_FULLSCREEN:
                    ToggleFullScreen ();
                    break;
                case ID_WINDOW_STRETCH:
                    GUI.Stretch = !GUI.Stretch;
                    RECT rect;
                    GetClientRect (GUI.hWnd, &rect);
                    InvalidateRect (GUI.hWnd, &rect, true);
                    break;
                case ID_FILE_SAVE_SPC_DATA:
					spc_is_dumping = 1;
//                    S9xSPCDump (S9xGetFilenameInc (".spc"));
                    break;
                case ID_FILE_SAVE_SRAM_DATA:
                    Memory.SaveSRAM (S9xGetFilename (".srm"));
                    break;
                case ID_FILE_RESET:
#ifdef NETPLAY_SUPPORT
                    if (Settings.NetPlayServer)
                    {
                        S9xNPReset ();
                        SetupSound (Settings.SoundPlaybackRate, 
                                    Settings.SixteenBitSound, 
                                    Settings.Stereo);
                    }
                    else
                    if (!Settings.NetPlay)
#endif
                    {
                        S9xReset ();
                        SetupSound (Settings.SoundPlaybackRate, 
                                    Settings.SixteenBitSound, 
                                    Settings.Stereo);
                    }
                    Settings.Paused = false;
                    break;
                case ID_FILE_PAUSE:
                    Settings.Paused = !Settings.Paused;
                    break;
                case ID_FILE_LOAD1:
                    FreezeUnfreeze (0, FALSE);
                    break;
                case ID_FILE_LOAD2:
                    FreezeUnfreeze (1, FALSE);
                    break;
                case ID_FILE_LOAD3:
                    FreezeUnfreeze (2, FALSE);
                    break;
                case ID_FILE_LOAD4:
                    FreezeUnfreeze (3, FALSE);
                    break;
                case ID_FILE_LOAD5:
                    FreezeUnfreeze (4, FALSE);
                    break;
                case ID_FILE_LOAD6:
                    FreezeUnfreeze (5, FALSE);
                    break;
                case ID_FILE_LOAD7:
                    FreezeUnfreeze (6, FALSE);
                    break;
                case ID_FILE_LOAD8:
                    FreezeUnfreeze (7, FALSE);
                    break;
                case ID_FILE_LOAD9:
                    FreezeUnfreeze (8, FALSE);
                    break;
                case ID_FILE_SAVE1:
                    FreezeUnfreeze (0, TRUE);
                    break;
                case ID_FILE_SAVE2:
                    FreezeUnfreeze (1, TRUE);
                    break;
                case ID_FILE_SAVE3:
                    FreezeUnfreeze (2, TRUE);
                    break;
                case ID_FILE_SAVE4:
                    FreezeUnfreeze (3, TRUE);
                    break;
                case ID_FILE_SAVE5:
                    FreezeUnfreeze (4, TRUE);
                    break;
                case ID_FILE_SAVE6:
                    FreezeUnfreeze (5, TRUE);
                    break;
                case ID_FILE_SAVE7:
                    FreezeUnfreeze (6, TRUE);
                    break;
                case ID_FILE_SAVE8:
                    FreezeUnfreeze (7, TRUE);
                    break;
                case ID_FILE_SAVE9:
                    FreezeUnfreeze (8, TRUE);
                    break;
                case ID_CHEAT_ENTER:
                    RestoreGUIDisplay ();
                    S9xRemoveCheats ();
                    _CheatCodeEntry (&Cheat);
                    S9xSaveCheatFile (S9xGetFilename (".cht"));
                    S9xApplyCheats ();
                    RestoreSNESDisplay ();
                    break;
                case ID_CHEAT_SEARCH:
                    RestoreGUIDisplay ();
                    _SearchForCheats (&Cheat);
                    S9xSaveCheatFile (S9xGetFilename (".cht"));
                    RestoreSNESDisplay ();
                    break;
                case ID_CHEAT_APPLY:
                    Settings.ApplyCheats = !Settings.ApplyCheats;
                    if (!Settings.ApplyCheats)
                        S9xRemoveCheats ();
                    else
                        S9xApplyCheats ();
                    break;
                case ID_OPTIONS_SETTINGS:
                    RestoreGUIDisplay ();
                    _SettingsDialog (&Settings, &GUI);
                    RestoreSNESDisplay ();
                    break;
                case ID_HELP_ABOUT:
                    RestoreGUIDisplay ();
                    _AboutDialog (VERSION);
                    RestoreSNESDisplay ();
                    break;
#ifdef DEBUGGER
                case ID_DEBUG_TRACE:
                {
                    Trace ();
                    break;
                }
                case ID_DEBUG_FRAME_ADVANCE:
                    CPU.Flags |= FRAME_ADVANCE_FLAG;
                    ICPU.FrameAdvanceCount = 1;
                    Settings.Paused = FALSE;
                    break;
#endif
				case IDM_7110_CACHE:
					RestoreGUIDisplay ();
                    DialogBox(g_hInst, MAKEINTRESOURCE(IDD_SPC7110_CACHE), hWnd, SP7PackConfig);
					RestoreSNESDisplay ();
                    break;
				case IDM_LOG_7110:
					Do7110Logging();
					break;
				case IDM_ROM_INFO:
					RestoreGUIDisplay ();
                    DialogBox(g_hInst, MAKEINTRESOURCE(IDD_ROM_INFO), hWnd, InfoProc);
					RestoreSNESDisplay ();
                    break;
                default:
                    if ((wParam & 0xffff) >= 0xFF00)
                    {
                        int i = (wParam & 0xffff) - 0xFF00;
                        int j = 0;
                        if (GUI.RecentGames)
                        {
                            while (GUI.RecentGames [j] && j != i)
                                j++;
                            if (i == j)
                            {
#ifdef NETPLAY_SUPPORT
                                if (Settings.NetPlay && !Settings.NetPlayServer)
                                {
                                    S9xMessage (S9X_INFO, S9X_NETPLAY_NOT_SERVER,
                                                "Disconnect from the NetPlay server first.");
                                    break;
                                }
#endif
                                if (!Settings.StopEmulation)
                                {
                                    Memory.SaveSRAM (S9xGetFilename (".srm"));
                                    S9xSaveCheatFile (S9xGetFilename (".cht"));
                                }
                                Settings.StopEmulation = !LoadROM (GUI.RecentGames [i]);
                                if (!Settings.StopEmulation) 
                                {
                                    Memory.LoadSRAM (S9xGetFilename (".srm"));
                                    S9xLoadCheatFile (S9xGetFilename (".cht"));
                                    S9xAddToRecentGames (GUI.RecentGames [i]);
                                    CheckDirectoryIsWritable (S9xGetFilename (".---"));
                                    CheckMenuStates ();
#ifdef NETPLAY_SUPPORT
                                    if (NPServer.SendROMImageOnConnect)
                                        S9xNPServerQueueSendingROMImage ();
                                    else
                                        S9xNPServerQueueSendingLoadROMRequest (Memory.ROMName);
#endif
                                    Settings.Paused = false;
                                }
                            }
                        }
                    }
                    break;
            }
            break;

	case WM_EXITMENULOOP:
            UpdateBackBuffer();
            S9xClearPause (PAUSE_MENU);
            break;

	case WM_ENTERMENULOOP:
            S9xSetPause (PAUSE_MENU);
#ifdef USE_GLIDE
            S9xGlideEnable (FALSE);
#endif
            CheckMenuStates ();

            SwitchToGDI();
            DrawMenuBar( GUI.hWnd);
            break;

	case WM_DESTROY:
			Memory.SaveSRAM(S9xGetFilename(".srm"));
			if(CleanUp7110)
				(*CleanUp7110)();
            PostQuitMessage (0);
            return (0);
        case WM_PAINT:
        {
            PAINTSTRUCT paint;

            BeginPaint (GUI.hWnd, &paint);
            EndPaint (GUI.hWnd, &paint);
            break;
        }
        case WM_SYSCOMMAND:
        {
            // Prevent screen saver from starting if not paused
            if ((Settings.ForcedPause || Settings.StopEmulation ||
                 Settings.Paused) &&
                (wParam == SC_SCREENSAVE || wParam == SC_MONITORPOWER))
                return (0);
            break;
        }
        case WM_ACTIVATE:
            if (LOWORD(wParam) == WA_INACTIVE)
            {
#ifdef USE_GLIDE
                if (VOODOO_MODE)
                {
                    S9xGlideEnable (FALSE);
#if 0
                    MoveWindow (GUI.hWnd, GUI.window_size.left,
                                GUI.window_size.top, 
                                GUI.window_size.right - GUI.window_size.left,
                                GUI.window_size.bottom - GUI.window_size.top,
                                TRUE);
#endif
                }
#endif
                S9xSetPause (PAUSE_INACTIVE_WINDOW);
            }
            else
            {
                S9xClearPause (PAUSE_INACTIVE_WINDOW);
                IPPU.ColorsChanged = TRUE;
            }
            break;
        case WM_QUERYNEWPALETTE:
//            if (!GUI.FullScreen && GUI.ScreenDepth == 8)
//                RealizePalette (GUI.WindowDC);
            break;
        case WM_SIZE:
            if (wParam == SIZE_RESTORED)
            {
                S9xClearPause (PAUSE_WINDOW_ICONISED);
            }
            if (wParam == SIZE_MINIMIZED || wParam == SIZE_MAXHIDE)
            {
#ifdef USE_GLIDE
                S9xGlideEnable (FALSE);
#endif
                S9xClearPause (PAUSE_WINDOW_ICONISED);
            }
#ifdef USE_DIRECTX3D
            D3DSizeChange (GUI.hWnd, wParam);
#endif
            break;
        case WM_MOVE:
            if (!VOODOO_MODE && !GUI.FullScreen && !Settings.ForcedPause)
            {
                GetWindowRect (GUI.hWnd, &GUI.window_size);
            }
#ifdef USE_DIRECTX3D            
            D3DMove (GUI.hWnd, lParam);
#endif

            break;
        case WM_DISPLAYCHANGE:
            if (!GUI.FullScreen)
            {
                if (!VOODOO_MODE && !OPENGL_MODE &&
                    DirectX.SetDisplayMode (GUI.Width, GUI.Height, GUI.Depth,
                                            !GUI.FullScreen, GUI.DoubleBuffered))
                {
                    S9xGraphicsDeinit();
                    S9xSetWinPixelFormat ();
                    S9xGraphicsInit();
                }
            }
            break;
        case WM_MOUSEMOVE:
            // Lo-word of lparam is xpos, hi-word is ypos
            if (!GUI.IgnoreNextMouseMove)
            {
                POINT p;
                p.x = (int16) (lParam & 0xffff);
                p.y = (int16) ((lParam >> 16) & 0xffff);
                ClientToScreen (GUI.hWnd, &p);
                if ((!Settings.ForcedPause && !Settings.StopEmulation && 
                     !Settings.Paused) &&
                    (IPPU.Controller == SNES_MOUSE ||
                     IPPU.Controller == SNES_MOUSE_SWAPPED))
                {
                    POINT middle;
                    RECT size;
                    
                    GetClientRect (GUI.hWnd, &size);
                    middle.x = (size.right - size.left) >> 1;
                    middle.y = (size.bottom - size.top) >> 1;
                    ClientToScreen (GUI.hWnd, &middle);
                    GUI.MouseX += p.x - middle.x;
                    GUI.MouseY += p.y - middle.y;
                    SetCursorPos (middle.x, middle.y);
                    GUI.IgnoreNextMouseMove = true;
                }
                else
                {
                    GUI.MouseX = p.x;
                    GUI.MouseY = p.y;
                }
            }
            else
                GUI.IgnoreNextMouseMove = false;

            S9xMouseOn ();
            break;
        case WM_LBUTTONDOWN:
            S9xMouseOn ();
            GUI.MouseButtons |= 1;
            break;
        case WM_LBUTTONUP:
            S9xMouseOn ();
            GUI.MouseButtons &= ~1;
            break;
        case WM_RBUTTONDOWN:
            S9xMouseOn ();
            GUI.MouseButtons |= 2;
            break;
        case WM_RBUTTONUP:
            S9xMouseOn ();
            GUI.MouseButtons &= ~2;
            break;
        case WM_MBUTTONDOWN:
            S9xMouseOn ();
            GUI.MouseButtons |= 4;
            break;
        case WM_MBUTTONUP:
            S9xMouseOn ();
            GUI.MouseButtons &= ~4;
            break;
        case WM_USER + 3:
            NetPlay.Answer = S9xLoadROMImage ((const char *) lParam);
            SetEvent (NetPlay.ReplyEvent);
            break;
        case WM_USER + 2:
            S9xMessage (0, 0, NetPlay.WarningMsg);
            break;
        case WM_USER + 1:
            RestoreGUIDisplay ();
            S9xRestoreWindowTitle ();
            MessageBox (GUI.hWnd, NetPlay.ErrorMsg, 
                        "Snes9X NetPlay Error", MB_OK | MB_ICONSTOP);
            RestoreSNESDisplay ();
            break;
        case WM_USER:
            if (NetPlay.ActionMsg [0] == 0)
                S9xRestoreWindowTitle ();
            else
            {
                char buf [NP_MAX_ACTION_LEN + 10];

                sprintf (buf, "%s %3d%%", NetPlay.ActionMsg, (int) lParam);
                SetWindowText (GUI.hWnd, buf);
            }
#if 0
            if ((int) lParam >= 0)
            {
                RestoreGUIDisplay ();
                _ProgressDialog (NetPlay.CurrentAction, (int) lParam);
            }
            else
            {
                _ProgressDialog ("", (int) lParam);
                RestoreSNESDisplay ();
            }
#endif
            break;
    }
    return DefWindowProc (hWnd, uMsg, wParam, lParam);
}

/*****************************************************************************/
/* WinInit                                                                   */
/*****************************************************************************/
BOOL WinInit( HINSTANCE hInstance)
{
    WNDCLASS wndclass;

    wndclass.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
    wndclass.lpfnWndProc = WinProc;
    wndclass.cbClsExtra = 0;
    wndclass.cbWndExtra = 0;
    wndclass.hInstance = hInstance;
    wndclass.hIcon = LoadIcon (NULL, "IDI_ICON1");
    wndclass.hCursor = NULL; //LoadCursor (NULL, IDC_ARROW);
    wndclass.hbrBackground = (HBRUSH)GetStockObject (BLACK_BRUSH);
    wndclass.lpszMenuName = NULL;
    wndclass.lpszClassName = "Snes9X: WndClass";
    
    GUI.hInstance = hInstance;
    
    if (!RegisterClass (&wndclass))
        return FALSE;
    
    if ((GUI.hMenu = LoadMenu (hInstance, MAKEINTRESOURCE( Languages[ GUI.Language].idMenu))) == NULL)
        return FALSE;
    
    char buf [100];
    sprintf (buf, "Snes9X v%s for Windows 95, NT 4.0, and higher", VERSION);

    DWORD dwExStyle;
    DWORD dwStyle;
    RECT rect;

    rect.left = rect.top = 0;
    rect.right = MAX_SNES_WIDTH;
    rect.bottom = MAX_SNES_HEIGHT;
    dwExStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
    dwStyle = WS_OVERLAPPEDWINDOW;

    AdjustWindowRectEx (&rect, dwStyle, FALSE, dwExStyle);
    if ((GUI.hWnd = CreateWindowEx (
        dwExStyle,
        "Snes9X: WndClass",
        buf,
        WS_CLIPSIBLINGS | 
        WS_CLIPCHILDREN |
        dwStyle,
        0, 0,
        rect.right - rect.left, rect.bottom - rect.top,
        NULL,
        GUI.hMenu,
        hInstance,
        NULL)) == NULL)
        return FALSE;
    
    GUI.hDC = GetDC (GUI.hWnd);

    GUI.Blank = LoadCursor (hInstance, MAKEINTRESOURCE (IDC_CURSOR_BLANK));
    GUI.GunSight = LoadCursor (hInstance, MAKEINTRESOURCE (IDC_CURSOR_SCOPE));
    GUI.Arrow = LoadCursor (NULL, IDC_ARROW);
    GUI.Accelerators = LoadAccelerators (hInstance, MAKEINTRESOURCE (IDR_SNES9X_ACCELERATORS));
    Settings.ForcedPause = 0;
    Settings.StopEmulation = TRUE;
    Settings.Paused = FALSE;
    
    return TRUE;
}

static inline char *SkipSpaces (char *p)
{
    while (*p && isspace (*p))
        p++;

    return (p);
}

void WinParseCommandLine (char *line)
{
    // Break the command line up into an array of string pointers, each pointer
    // points at a separate word or character sequence enclosed in quotes.

#define MAX_PARAMETERS 100
    char *p = line;
    static char *parameters [MAX_PARAMETERS];
    int count = 0;

    parameters [count++] = "Snes9XW";

    while (count < MAX_PARAMETERS && *p)
    {
        p = SkipSpaces (p);
        if (*p == '"')
        {
            p++;
            parameters [count++] = p;
            while (*p && *p != '"')
                p++;
            *p++ = 0;
        }
        else
        if (*p == '\'')
        {
            p++;
            parameters [count++] = p;
            while (*p && *p != '\'')
                p++;
            *p++ = 0;
        }
        else
        {
            parameters [count++] = p;
            while (*p && !isspace (*p))
                p++;
            if (!*p)
                break;
            *p++ = 0;
        }
    }
    rom_filename = S9xParseArgs (parameters, count);
}

void S9xParseArg (char **argv, int &i, int argc)
{
    if (strcasecmp (argv [i], "-restore") == 0)
    {
        WinDeleteRegistryEntries ();
        WinSetDefaultValues ();
    }
}

void S9xExtraUsage ()
{
}

/*****************************************************************************/
/* WinSave/WinLoad - Save/Load the settings to/from the registry             */
/*****************************************************************************/
#define SetKey( n, v)\
	RegSetValueEx( hKey, n, 0, REG_BINARY, (const unsigned char *) &v, sizeof( v));
#define GetKey( n, v)\
	cbData = sizeof( v);\
	RegQueryValueEx( hKey, n, 0, NULL, (unsigned char *) &v, &cbData);
#define GetKeyDef( n, v, d)\
	cbData = sizeof( v);\
	if( RegQueryValueEx( hKey, n, 0, NULL, (unsigned char *) &v, &cbData) != ERROR_SUCCESS)\
		v = d;

void WinDeleteRegKey (HKEY hKeyRoot, const char *key)
{
    FILETIME last_update;
    DWORD count;
    HKEY hKey;
    char name_buf [100];
    char class_buf [200];
    DWORD name_len;
    DWORD class_len;

    if (RegOpenKeyEx (hKeyRoot, key, 0, KEY_READ | KEY_WRITE, &hKey) == ERROR_SUCCESS)
    {
        bool8 found = TRUE;

        count = 0;
        RegQueryInfoKey (hKey, NULL, NULL, NULL, &count,
                         NULL, NULL, NULL, NULL, NULL, NULL, NULL);
        
        for (; found && count != 0; count--)
        {
            name_len = sizeof (name_buf);
            class_len = sizeof (class_buf);
            if (RegEnumKeyEx (hKey, count - 1, name_buf, &name_len, NULL, 
                              class_buf, &class_len, &last_update) == ERROR_SUCCESS)
            {
                found = TRUE;
                WinDeleteRegKey (hKey, name_buf);
                RegDeleteKey (hKey, name_buf);
            }
            else
                found = FALSE;
        }
    }
}

static void WinDeleteRegistryEntries ()
{
    WinDeleteRegKey (HKEY_CURRENT_USER, S9X_REG_KEY_BASE);
}

void WinSave( void)
{
    HKEY hKey;
    char tmp [100];

    if (RegCreateKeyEx (HKEY_CURRENT_USER, 
                        S9X_REG_KEY_BASE "\\" S9X_REG_KEY_VERSION, 0, NULL, 
                        REG_OPTION_NON_VOLATILE, KEY_ALL_ACCESS, NULL, &hKey,
                        NULL) == ERROR_SUCCESS)
    {
        SetKey( "Language", GUI.Language);
        SetKey( "Playback Rate", Settings.SoundPlaybackRate);
        SetKey( "Stereo Sound", Settings.Stereo);
        SetKey( "16bit Sound", Settings.SixteenBitSound);
        SetKey( "Interpolated Sound", Settings.InterpolatedSound);
        SetKey( "Sync Sound", Settings.SoundSync);
        SetKey( "Sound Envelope Height Reading", Settings.SoundEnvelopeHeightReading);
        SetKey( "Reverse Stereo", Settings.ReverseStereo);
        SetKey( "Disable Echo", Settings.DisableSoundEcho);
        SetKey( "Disable Sample Caching", Settings.DisableSampleCaching);
        SetKey( "Disable Master Volume", Settings.DisableMasterVolume);
        SetKey( "Mute Sound", Settings.Mute);
        SetKey( "Alt Sample Code Method", Settings.AltSampleDecode);
        SetKey( "Enable SPC700", Settings.NextAPUEnabled);
        SetKey( "Sound CPU Skip Method", Settings.SoundSkipMethod);
        SetKey( "Paused Frames Before Muting Sound", GUI.PausedFramesBeforeMutingSound);
        SetKey( "Sound Buffer Length", Settings.SoundBufferSize);
        SetKey( "Sound Mix Interval", Settings.SoundMixInterval);
        SetKey( "Sound Driver", Settings.SoundDriver);

        SetKey( "Width", GUI.Width);
        SetKey( "Height", GUI.Height);
        SetKey( "Depth", GUI.Depth);
        SetKey( "Scale", GUI.NextScale);
        SetKey( "DoubleBuffered", GUI.DoubleBuffered);
        SetKey( "Show FPS", Settings.DisplayFrameRate);

        SetKey( "Render SixteenBit", Settings.SixteenBit);
        SetKey( "Enable Transparency", Settings.Transparency);
        SetKey( "Support HiRes", Settings.SupportHiRes);
        SetKey( "Skip Frames", Settings.SkipFrames);
        SetKey( "Turbo Skip Frames", Settings.TurboSkipFrames);
        SetKey( "Turbo Mode Toggle", GUI.TurboModeToggle);
        SetKey( "Auto Max Skip Frames", Settings.AutoMaxSkipFrames);

        SetKey( "Joypad #1", Joypad[0]);
        SetKey( "Joypad #2", Joypad[1]);
        SetKey( "Joypad #3", Joypad[2]);
        SetKey( "Joypad #4", Joypad[3]);
        SetKey( "Joypad #5", Joypad[4]);
        SetKey( "Window Geometry", GUI.window_size);
        SetKey( "Stretch", GUI.Stretch);
        SetKey( "Fullscreen", GUI.FullScreen);
        RegSetValueEx (hKey, "Freeze File Directory", 0, REG_SZ,
                       (const unsigned char *) GUI.FreezeFileDir, 
                       strlen (GUI.FreezeFileDir) + 1);
        SetKey( "Apply Cheats", Settings.ApplyCheats);
        SetKey( "Auto Save Delay", Settings.AutoSaveDelay);

        SetKey( "NetPlay Port", Settings.Port);
        SetKey( "NetPlay Sync By Reset", NPServer.SyncByReset);
        SetKey( "NetPlay Send ROM On Connect", NPServer.SendROMImageOnConnect);
        SetKey( "NetPlay Max Frame Skip", NetPlay.MaxFrameSkip);
        SetKey( "NetPlay Max Behind Frame Count", NetPlay.MaxBehindFrameCount);
        SetKey( "Bi-Linear Filter Mode7", Settings.Mode7Interpolate);

		DWORD mk_temp;
		if(LoadUp7110==&SPC7110Open)
			mk_temp=1;
		else if(LoadUp7110==&SPC7110Grab)
			mk_temp=2;
		else mk_temp=0;
			SetKey( "SPC7110 Cache", mk_temp);

		SetKey ( "SPC7110 Size", cacheMegs);

        if (GUI.RecentGames)
        {
            int i = 0;

            while (GUI.RecentGames [i])
            {
                sprintf (tmp, "Recent Game #%d", i + 1);
                RegSetValueEx (hKey, tmp, 0, REG_SZ,
                       (const unsigned char *) GUI.RecentGames [i],
                       strlen (GUI.RecentGames [i]) + 1);
                i++;
            }
        }
        RegCloseKey( hKey);
    }
}

static void WinSetDefaultValues ()
{
    GUI.IgnoreNextMouseMove = false;
    GUI.window_size.left = 0;
    GUI.window_size.right = 264 + 256;
    GUI.window_size.top = 0;
    GUI.window_size.bottom = 285 + 239;
    GUI.Width = 640;
    GUI.Height = 480;
    GUI.Depth = 16;
    GUI.Scale = 0;
    GUI.NextScale = 0;
    GUI.DoubleBuffered = false;
    GUI.FullScreen = false;
    GUI.Stretch = false;
    GUI.PausedFramesBeforeMutingSound = 20;
    GUI.FlipCounter = 0;
    GUI.NumFlipFrames = 1;
    WinDeleteRecentGamesList ();

    // ROM Options
    memset (&Settings, 0, sizeof (Settings));

    Settings.ForceLoROM = false;
    Settings.ForceInterleaved = false;

    Settings.ForceNotInterleaved = false;
    Settings.ForceInterleaved = false;
    Settings.ForceInterleaved2 = false;

    Settings.ForcePAL = false;
    Settings.ForceNTSC = false;
    Settings.ForceHeader = false;
    Settings.ForceNoHeader = false;

    // Sound options
    Settings.SoundSync = FALSE;
    Settings.InterpolatedSound = TRUE;
    Settings.SoundEnvelopeHeightReading = TRUE;
    Settings.DisableSoundEcho = FALSE;
    Settings.DisableMasterVolume = FALSE;
    Settings.Mute = FALSE;
    Settings.SoundSkipMethod = 0;
    Settings.SoundPlaybackRate = 22050;
    Settings.SixteenBitSound = TRUE;
    Settings.Stereo = TRUE;
    Settings.AltSampleDecode = FALSE;
    Settings.ReverseStereo = FALSE;
    Settings.SoundDriver = WIN_SNES9X_DIRECT_SOUND_DRIVER;
    Settings.SoundBufferSize = 1;
    Settings.SoundMixInterval = 10;

    // Tracing options
    Settings.TraceDMA = false;
    Settings.TraceHDMA = false;
    Settings.TraceVRAM = false;
    Settings.TraceUnknownRegisters = false;
    Settings.TraceDSP = false;

    // Joystick options
    Settings.SwapJoypads = false;
    Settings.JoystickEnabled = false;

    // ROM timing options (see also H_Max above)
    Settings.PAL = false;
    Settings.FrameTimePAL = 20;
    Settings.FrameTimeNTSC = 17;
    Settings.FrameTime = 17;

    // CPU options 
    Settings.CyclesPercentage = 100;
    Settings.Shutdown = true;
    Settings.ShutdownMaster = true;
    Settings.NextAPUEnabled = Settings.APUEnabled = TRUE;
    Settings.DisableIRQ = false;
    Settings.Paused = false;
    Settings.H_Max = SNES_CYCLES_PER_SCANLINE;
    Settings.HBlankStart = (256 * Settings.H_Max) / SNES_HCOUNTER_MAX;
    Settings.SkipFrames = AUTO_FRAMERATE;

    // ROM image and peripheral options
    Settings.ForceSuperFX = false;
    Settings.ForceNoSuperFX = false;
    Settings.MultiPlayer5 = true;
    Settings.Mouse = true;
    Settings.SuperScope = true;
    Settings.MultiPlayer5Master = true;
    Settings.SuperScopeMaster = true;
    Settings.MouseMaster = true;
    Settings.SuperFX = false;
 
    // SNES graphics options
    Settings.BGLayering = false;
    Settings.DisableGraphicWindows = false;
    Settings.ForceTransparency = false;
    Settings.ForceNoTransparency = false;
    Settings.DisableHDMA = false;
    Settings.Mode7Interpolate = false;
    Settings.DisplayFrameRate = false;
    Settings.SixteenBit = true;
    Settings.Transparency = true;
    Settings.SupportHiRes = true;

    Settings.AutoSaveDelay = 0;
    Settings.ApplyCheats = true;

    Settings.TurboMode = false;
    Settings.TurboSkipFrames = 15;
    GUI.TurboModeToggle = false;
    Settings.AutoMaxSkipFrames = 10;

    Settings.Port = 1996;
    NetPlay.MaxFrameSkip = 10;
    NetPlay.MaxBehindFrameCount = 10;
    NPServer.SyncByReset = true;
    NPServer.SendROMImageOnConnect = false;

    GUI.FreezeFileDir [0] = 0;
}

void WinLoad( void)
{
    HKEY hKey;
    unsigned long cbData;
    char tmp [100];
    char path [_MAX_PATH + 1];

    WinSetDefaultValues ();

    if (RegOpenKeyEx (HKEY_CURRENT_USER, 
                      S9X_REG_KEY_BASE "\\" S9X_REG_KEY_VERSION,
                      0, KEY_READ, &hKey) == ERROR_SUCCESS)
    {
        GetKey( "Language", GUI.Language);
        GetKey( "Stereo Sound", Settings.Stereo);

        GetKeyDef( "Width", GUI.Width, 640);
        GetKeyDef( "Height", GUI.Height, 480);
        GetKeyDef( "Depth", GUI.Depth, 16);
        GetKeyDef( "Scale", GUI.Scale, 0);
        GUI.NextScale = GUI.Scale;
        GetKeyDef( "DoubleBuffered", GUI.DoubleBuffered, false);
        GetKeyDef( "Show FPS", Settings.DisplayFrameRate, false);

        GetKeyDef( "Render SixteenBit", Settings.SixteenBit, true);
        GetKeyDef( "Enable Transparency", Settings.Transparency, true);
        GetKeyDef( "Support HiRes", Settings.SupportHiRes, true);
        GetKeyDef( "Skip Frames", Settings.SkipFrames, AUTO_FRAMERATE);
        GetKeyDef( "Turbo Skip Frames", Settings.TurboSkipFrames, 15);
        GetKeyDef( "Turbo Mode Toggle", GUI.TurboModeToggle, FALSE);
        GetKeyDef( "Auto Max Skip Frames", Settings.AutoMaxSkipFrames, 10);
        GetKeyDef( "Bi-Linear Filter Mode7", Settings.Mode7Interpolate, false);

        GetKey( "Joypad #1", Joypad[0]);
        GetKey( "Joypad #2", Joypad[1]);
        GetKey( "Joypad #3", Joypad[2]);
        GetKey( "Joypad #4", Joypad[3]);
        GetKey( "Joypad #5", Joypad[4]);
        GetKey( "Window Geometry", GUI.window_size);
        GetKeyDef( "Stretch", GUI.Stretch, false);
        GetKeyDef( "Fullscreen", GUI.FullScreen, false);
        GetKeyDef( "Playback Rate", Settings.SoundPlaybackRate, 22050);
        GetKeyDef( "Stereo Sound", Settings.Stereo, true);
        GetKeyDef( "16bit Sound", Settings.SixteenBitSound, true);
        GetKeyDef( "Interpolated Sound", Settings.InterpolatedSound, true);
        GetKeyDef( "Sync Sound", Settings.SoundSync, false);
        GetKeyDef( "Sound Envelope Height Reading", Settings.SoundEnvelopeHeightReading, true);
        GetKeyDef( "Reverse Stereo", Settings.ReverseStereo, false);
        GetKeyDef( "Disable Echo", Settings.DisableSoundEcho, false);
        GetKeyDef( "Disable Sample Caching", Settings.DisableSampleCaching, false);
        GetKeyDef( "Disable Master Volume", Settings.DisableMasterVolume, false);
        GetKeyDef( "Mute Sound", Settings.Mute, false);
        GetKeyDef( "Alt Sample Code Method", Settings.AltSampleDecode, false);
        GetKeyDef( "Enable SPC700", Settings.NextAPUEnabled, true);
        Settings.APUEnabled = Settings.NextAPUEnabled;
        GetKeyDef( "Sound CPU Skip Method", Settings.SoundSkipMethod, 0);
        GetKeyDef( "Paused Frames Before Muting Sound", GUI.PausedFramesBeforeMutingSound, 20);
        GetKeyDef( "Sound Buffer Length", Settings.SoundBufferSize, 1);
        GetKeyDef( "Sound Mix Interval", Settings.SoundMixInterval, 10);
        GetKeyDef( "Sound Driver", Settings.SoundDriver, WIN_SNES9X_DIRECT_SOUND_DRIVER);
        GetKeyDef( "Auto Save Delay", Settings.AutoSaveDelay, 30);
        GetKeyDef( "Apply Cheats", Settings.ApplyCheats, true);
        
        GetKeyDef( "NetPlay Port", Settings.Port, 1996);
        GetKeyDef( "NetPlay Sync By Reset", NPServer.SyncByReset, true);
        GetKeyDef( "NetPlay Send ROM On Connect", NPServer.SendROMImageOnConnect, false);
        GetKeyDef( "NetPlay Max Frame Skip", NetPlay.MaxFrameSkip, 10);
        GetKeyDef( "NetPlay Max Behind Frame Count", NetPlay.MaxBehindFrameCount, 10);

        cbData = _MAX_PATH + 1;
        RegQueryValueEx (hKey, "Freeze File Directory", 0, NULL,
                         (unsigned char *) GUI.FreezeFileDir, &cbData);

		DWORD mk_temp;
		GetKeyDef( "SPC7110 Cache", mk_temp, 0);
		switch(mk_temp)
		{
			case 2:
				LoadUp7110=&SPC7110Grab;
				break;
			case 1:
				LoadUp7110=&SPC7110Open;
				break;
			case 0:
			default:LoadUp7110=&SPC7110Load;break;
		}
		GetKeyDef ( "SPC7110 Size", cacheMegs, 5);
        // Count how many recent games filenames are saved in the registry
        int i = -1;
        do
        {
            i++;
            sprintf (tmp, "Recent Game #%d", i + 1);
            cbData = _MAX_PATH + 1;
        } while (RegQueryValueEx (hKey, tmp, 0, NULL,
                                  (unsigned char *) path, &cbData) == ERROR_SUCCESS);
        if (i > 0)
        {
            // Now populate the RecentGames array.
            GUI.RecentGames = new char *[i + 1];
            GUI.RecentGames [i] = NULL;
            for (int j = 0; j < i; j++)
            {
                sprintf (tmp, "Recent Game #%d", j + 1);
                cbData = _MAX_PATH + 1;
                RegQueryValueEx (hKey, tmp, 0, NULL,
                                 (unsigned char *) path, &cbData);
                GUI.RecentGames [j] = strdup (path);
            }
        }
        RegCloseKey( hKey);
    } 

    GUI.ScreenCleared = true;
}

extern long _rate;
extern bool _sixteen_bit;
extern bool _stereo;
extern long _buffernos;
extern long _buffersize;
extern long _samplecount;
extern long _bytecount;
extern unsigned long _interval;

volatile unsigned long LastSound = 0;
uint32 next_frame_time = 0;

VOID CALLBACK SoundTimer (UINT idEvent, UINT uMsg, DWORD dwUser, 
                          DWORD dw1, DWORD dw2)
{
    void ProcessSound (void);

    if (Settings.SoundDriver == WIN_SNES9X_DIRECT_SOUND_DRIVER)
        ProcessSound ();
}

VOID CALLBACK FrameTimer( UINT idEvent, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2)
{
    if (DirectX.FrameCount == DirectX.LastFrameCount)
        DirectX.IdleCount++;
    else
    {
        DirectX.IdleCount = 0;
        DirectX.LastFrameCount = DirectX.FrameCount;
    }

#ifdef NETPLAY_SUPPORT
    if (Settings.NetPlay && !Settings.NetPlayServer)
        return;
#endif

#ifdef NETPLAY_SUPPORT
    if (Settings.NetPlayServer)
    {
        if (Settings.Paused || Settings.StopEmulation || Settings.ForcedPause)
        {
            WaitForSingleObject (GUI.ServerTimerSemaphore, 0);
            return;
        }
        ReleaseSemaphore (GUI.ServerTimerSemaphore, 1, NULL);

        if (Settings.NetPlay)
            return;
    }
    else
#endif
    if (Settings.SkipFrames != AUTO_FRAMERATE || Settings.TurboMode ||
        Settings.Paused || Settings.StopEmulation || Settings.ForcedPause)
    {
        WaitForSingleObject (GUI.FrameTimerSemaphore, 0);
        return;
    }
    ReleaseSemaphore (GUI.FrameTimerSemaphore, 1, NULL);
}

/*****************************************************************************/
/* WinMain                                                                   */
/*****************************************************************************/
extern "C" void S9xMainLoop(void);
int Init3d (HWND);
extern "C" void DeinitS9x(void);
int WINAPI WinMain( 
	HINSTANCE hInstance,
	HINSTANCE hPrevInstance,
	LPSTR lpCmdLine,
	int nCmdShow)
{
    DWORD wSoundTimerRes;

    GUI.RecentGames = NULL;
    WinLoad();
    WinParseCommandLine (lpCmdLine);

    WinInit (hInstance);
//    GUI.hWnd = _MainWindow ();
//    SetMenu (GUI.hWnd, GUI.hMenu);
    _InitDLL (GUI.hWnd);
    
    if (_GUIVersion () != GUI_VERSION)
    {
        MessageBox (GUI.hWnd, "\
Incorrect version of Snes9XW.dll dynamic-link library file found.\n\n\
Please check you extracted both the Snes9XW.exe and Snes9XW.dll\n\
files into the same directory from the Snes9XW distribution\n\
zip file.", "Snes9X - Incorrect Snes9XW.dll Version", MB_OK | MB_ICONSTOP);
        return false; 
    }

    DirectX.InitDirectSound ();

    if (!SetupSound (Settings.SoundPlaybackRate, Settings.SixteenBitSound, 
                     Settings.Stereo))
    {	
        MessageBox( GUI.hWnd, Languages[ GUI.Language].errInitDS, "Snes9X - DirectSound", MB_OK | MB_ICONINFORMATION);
    }

#ifdef USE_GLIDE
    if (VOODOO_MODE)
    {
        if (!S9xVoodooInitialise ())
        {
            GUI.Scale = 0;
            GUI.NextScale = 0;
        }
        else
        {
            GUI.FullScreen = FALSE;
            Settings.SixteenBit = TRUE;
        }
    }
#endif

#ifdef USE_OPENGL
    if (OPENGL_MODE)
    {
        if (!S9xOpenGLInit ())
        {
            GUI.Scale = 0;
            GUI.NextScale = 0;
        }
    }
#endif

    if (!VOODOO_MODE && !OPENGL_MODE && !DirectX.InitDirectX ())
    {
        MessageBox (GUI.hWnd, Languages[ GUI.Language].errInitDD, "Snes9X - DirectX", MB_OK | MB_ICONSTOP);
        return false; 
    }
    
    if (!GUI.FullScreen)
    {
        MoveWindow (GUI.hWnd, GUI.window_size.left,
                    GUI.window_size.top, 
                    GUI.window_size.right - GUI.window_size.left,
                    GUI.window_size.bottom - GUI.window_size.top, TRUE);
    }

    if (!VOODOO_MODE && !OPENGL_MODE && 
        !DirectX.SetDisplayMode (GUI.Width, GUI.Height, GUI.Depth, 
                                 !GUI.FullScreen, GUI.DoubleBuffered))
    {
        MessageBox( GUI.hWnd, Languages[ GUI.Language].errModeDD, "Snes9X - DirectDraw(7)", MB_OK | MB_ICONSTOP);
        GUI.FullScreen = FALSE;
        if (!DirectX.SetDisplayMode (GUI.Width, GUI.Height, GUI.Depth, 
                                     !GUI.FullScreen, GUI.DoubleBuffered))
            return (false);
    }

    if (!GUI.FullScreen)
    {
        RECT rect;
        GetClientRect (GUI.hWnd, &rect);
        InvalidateRect (GUI.hWnd, &rect, true);
    }

    void InitSnes9X( void);
    S9xSetRecentGames ();
    ShowWindow (GUI.hWnd, SW_SHOWNORMAL);
    SetForegroundWindow (GUI.hWnd);
    SetFocus (GUI.hWnd);

#ifdef USE_DIRECTX3D
    if (Init3d (GUI.hWnd) != -100)
    {
        MessageBox (GUI.hWnd, "InitD3D failed.", "InitD3D Error", MB_OK | MB_ICONINFORMATION);
        exit (1);
    }
#endif

    InitSnes9X ();

    TIMECAPS tc;
    if (timeGetDevCaps(&tc, sizeof(TIMECAPS))== TIMERR_NOERROR)
    {
        wSoundTimerRes = min(max(tc.wPeriodMin, 1), tc.wPeriodMax);
        timeBeginPeriod (wSoundTimerRes);
    }

    Settings.StopEmulation = true;
    GUI.hFrameTimer = timeSetEvent (20, 0, FrameTimer, 0, TIME_PERIODIC);
    GUI.hSoundTimer = timeSetEvent (5, 0, SoundTimer, 0, TIME_PERIODIC);

    GUI.FrameTimerSemaphore = CreateSemaphore (NULL, 0, 10, NULL);
    GUI.ServerTimerSemaphore = CreateSemaphore (NULL, 0, 10, NULL);

    if (GUI.hFrameTimer == NULL)
    {
        MessageBox( GUI.hWnd, Languages[ GUI.Language].errFrameTimer, "Snes9X - Frame Timer", MB_OK | MB_ICONINFORMATION);
    }

    Settings.StopEmulation = !rom_filename || !LoadROM (rom_filename);

    if (!Settings.StopEmulation)
    {
        Memory.LoadSRAM (S9xGetFilename (".srm"));
        S9xLoadCheatFile (S9xGetFilename (".cht"));
        CheckDirectoryIsWritable (S9xGetFilename (".---"));
        CheckMenuStates ();
    }

    if (!Settings.StopEmulation)
    {
        if (IPPU.Controller == SNES_SUPERSCOPE)
            SetCursor (GUI.GunSight);
        else
        {
            SetCursor (GUI.Arrow);
            GUI.CursorTimer = 60;
        }
    }

    MSG msg;

    while (TRUE)
    {
        // Changed 9/1/2001
        while (Settings.StopEmulation || Settings.Paused ||
               Settings.ForcedPause ||
               PeekMessage (&msg, NULL, 0, 0, PM_NOREMOVE))
        {
            if (!GetMessage (&msg, NULL, 0, 0))
                goto loop_exit;
            
            if (!TranslateAccelerator (GUI.hWnd, GUI.Accelerators, &msg))
            {
                TranslateMessage (&msg);
                DispatchMessage (&msg);
            }
        }

#ifdef NETPLAY_SUPPORT
        if (!Settings.NetPlay || !NetPlay.PendingWait4Sync ||
            WaitForSingleObject (GUI.ClientSemaphore, 100) != WAIT_TIMEOUT)
        {
            if (NetPlay.PendingWait4Sync)
            {
                NetPlay.PendingWait4Sync = FALSE;
                NetPlay.FrameCount++;
                S9xNPStepJoypadHistory ();
            }
#endif            
            S9xMainLoop();
            DirectX.FrameCount++;
#ifdef NETPLAY_SUPPORT
        }
#endif
        if (CPU.Flags & DEBUG_MODE_FLAG)
        {
            Settings.Paused = TRUE;
            CPU.Flags &= ~DEBUG_MODE_FLAG;
        }
        if (GUI.CursorTimer)
        {
            if (--GUI.CursorTimer == 0)
            {
                if (IPPU.Controller != SNES_SUPERSCOPE)
                    SetCursor (NULL);
            }
        }
    }

loop_exit:
#ifdef USE_GLIDE
    S9xGlideEnable (FALSE);
#endif

    if (GUI.hSoundTimer)
        timeKillEvent (GUI.hSoundTimer);

    if( GUI.hFrameTimer)
    {	
        timeKillEvent (GUI.hFrameTimer);
        timeEndPeriod (wSoundTimerRes);
    }

    if (!Settings.StopEmulation)
    {
        Memory.SaveSRAM (S9xGetFilename (".srm"));
        S9xSaveCheatFile (S9xGetFilename (".cht"));
    }
    if (!VOODOO_MODE && !GUI.FullScreen)
        GetWindowRect (GUI.hWnd, &GUI.window_size);

    if (OPENGL_MODE)
        S9xOpenGLDeinit ();

    WinSave ();
	Memory.Deinit();
	if(Settings.GlideEnable)
		S9xGlideDeinit();
	else if (Settings.OpenGLEnable)
		S9xOpenGLDeinit();
	S9xGraphicsDeinit();
	S9xDeinitAPU();
	WinDeleteRecentGamesList ();
	DeinitS9x();

	//uncomment to find memory leaks.
	//	_CrtDumpMemoryLeaks();
    return msg.wParam;
}

void RestoreGUIDisplay ()
{
    S9xSetPause (PAUSE_RESTORE_GUI);
#ifdef USE_GLIDE
    S9xGlideEnable (FALSE);
#endif
    if (!VOODOO_MODE && !OPENGL_MODE && GUI.FullScreen && 
        (GUI.Width < 640 || GUI.Height < 400) &&
        !DirectX.SetDisplayMode (640, 480, 0, !GUI.FullScreen, false))
    {
        MessageBox (GUI.hWnd, Languages[ GUI.Language].errModeDD, "Snes9X - DirectDraw(1)", MB_OK | MB_ICONSTOP);
        S9xClearPause (PAUSE_RESTORE_GUI);
        return;
    }
    SwitchToGDI();
    S9xClearPause (PAUSE_RESTORE_GUI);
}

void RestoreSNESDisplay ()
{
    if (!VOODOO_MODE && !OPENGL_MODE && 
        !DirectX.SetDisplayMode (GUI.Width, GUI.Height, GUI.Depth,
                                 !GUI.FullScreen, GUI.DoubleBuffered))
    {
        MessageBox (GUI.hWnd, Languages[ GUI.Language].errModeDD, "Snes9X - DirectDraw(4)", MB_OK | MB_ICONSTOP);
        return;
    }
#ifdef USE_GLIDE
    if (VOODOO_MODE && Glide.voodoo_present)
    {
        //S9xGlideEnable (TRUE);
    }
#endif
    
    UpdateBackBuffer();
}

void FreezeUnfreeze (int slot, bool8 freeze)
{
    static char filename [_MAX_PATH + 1];
    char drive [_MAX_DRIVE + 1];
    char dir [_MAX_DIR + 1];
    char fname [_MAX_FNAME + 1];
    char ext [_MAX_EXT + 1];
    
#ifdef NETPLAY_SUPPORT
    if (!freeze && Settings.NetPlay && !Settings.NetPlayServer)
    {
        S9xMessage (S9X_INFO, S9X_NETPLAY_NOT_SERVER,
                    "Only the server is allowed to load freeze files.");
        return;
    }
#endif

    _splitpath (Memory.ROMFilename, drive, dir, fname, ext);
    sprintf (ext, ".%03d", slot);
    if (GUI.FreezeFileDir [0])
    {
        strcpy (filename, GUI.FreezeFileDir);
        strcat (filename, "\\");
        strcat (filename, fname);
        strcat (filename, ext);
    }
    else
        _makepath (filename, drive, dir, fname, ext);
    
    if (!freeze && _access (filename, 0) != 0 && slot < 10)
    {
        static char *digits = "t123456789";
        sprintf (ext, ".zs%c", digits [slot]);
        if (GUI.FreezeFileDir [0])
        {
            strcpy (filename, GUI.FreezeFileDir);
            strcat (filename, "\\");
            strcat (filename, fname);
            strcat (filename, ext);
        }
        else
            _makepath (filename, drive, dir, fname, ext);
    }

    S9xSetPause (PAUSE_FREEZE_FILE);
    
    if (freeze)
        S9xFreezeGame (filename);
    else
    {
        if (S9xUnfreezeGame (filename))
        {
            S9xMessage (S9X_INFO, S9X_FREEZE_FILE_INFO, S9xBasename (filename));
#ifdef NETPLAY_SUPPORT
            S9xNPServerQueueSendingFreezeFile (filename);
#endif
            UpdateBackBuffer();
        }
    }
    
    S9xClearPause (PAUSE_FREEZE_FILE);
}

void CheckDirectoryIsWritable (const char *filename)
{
    FILE *fs = fopen (filename, "w+");

    if (fs == NULL)
        MessageBox (GUI.hWnd, "\
The folder where Snes9X saves emulated save RAM files and\n\
game save positions (freeze files) is currently set to a\n\
read-only folder.\n\n\
If you do not change the game save folder, Snes9X will be\n\
unable to save your progress in this game. Change the folder\n\
from the Settings Dialog available from the Options menu.\n\n\
The default the save folder is the same as the ROM image\n\
folder, if no value is set.\n",
                "Snes9X: Unable to save file warning", 
                MB_OK | MB_ICONINFORMATION);
    else
    {
        fclose (fs);
        remove (filename);
    }
}

static void CheckMenuStates ()
{
    MENUITEMINFO mii;
    int i;

    ZeroMemory( &mii, sizeof( mii));
    mii.cbSize = sizeof( mii);
    mii.fMask = MIIM_STATE;
    
    mii.fState = GUI.FullScreen ? MFS_CHECKED : MFS_UNCHECKED;
    SetMenuItemInfo (GUI.hMenu, ID_WINDOW_FULLSCREEN, FALSE, &mii);
    mii.fState = GUI.Stretch ? MFS_CHECKED : MFS_UNCHECKED;
    SetMenuItemInfo (GUI.hMenu, ID_WINDOW_STRETCH, FALSE, &mii);
    mii.fState = Settings.Paused ? MFS_CHECKED : MFS_UNCHECKED;
    SetMenuItemInfo (GUI.hMenu, ID_FILE_PAUSE, FALSE, &mii);
    
    mii.fState = MFS_UNCHECKED;
    if (Settings.StopEmulation)
        mii.fState |= MFS_DISABLED;
    SetMenuItemInfo (GUI.hMenu, ID_FILE_SAVE_SPC_DATA, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_SAVE_SRAM_DATA, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_SAVE1, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_SAVE2, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_SAVE3, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_SAVE4, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_SAVE5, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_SAVE6, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_SAVE7, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_SAVE8, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_SAVE9, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_LOAD1, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_LOAD2, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_LOAD3, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_LOAD4, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_LOAD5, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_LOAD6, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_LOAD7, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_LOAD8, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_LOAD9, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_PAUSE, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_FILE_RESET, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_CHEAT_ENTER, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_CHEAT_SEARCH, FALSE, &mii);
    
    if (Settings.NetPlay && !Settings.NetPlayServer)
        mii.fState = MFS_DISABLED;
    else
        mii.fState = Settings.NetPlayServer ? MFS_CHECKED : MFS_UNCHECKED;
    SetMenuItemInfo (GUI.hMenu, ID_NETPLAY_SERVER, FALSE, &mii);

    mii.fState = Settings.NetPlay && !Settings.NetPlayServer ? 0 : MFS_DISABLED;
    SetMenuItemInfo (GUI.hMenu, ID_NETPLAY_DISCONNECT, FALSE, &mii);

    mii.fState = Settings.NetPlay || Settings.NetPlayServer ? MFS_DISABLED : 0;
    SetMenuItemInfo (GUI.hMenu, ID_NETPLAY_CONNECT, FALSE, &mii);

    mii.fState = NPServer.SendROMImageOnConnect ? MFS_CHECKED : MFS_UNCHECKED;
    SetMenuItemInfo (GUI.hMenu, ID_NETPLAY_SEND_ROM_ON_CONNECT, FALSE, &mii);

    mii.fState = NPServer.SyncByReset ? MFS_CHECKED : MFS_UNCHECKED;
    SetMenuItemInfo (GUI.hMenu, ID_NETPLAY_SYNC_BY_RESET, FALSE, &mii);

    mii.fState = Settings.NetPlayServer ? 0 : MFS_DISABLED;
    SetMenuItemInfo (GUI.hMenu, ID_NETPLAY_SYNC, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_NETPLAY_ROM, FALSE, &mii);

    mii.fState = Settings.ApplyCheats ? MFS_CHECKED : MFS_UNCHECKED;
    if (Settings.StopEmulation)
        mii.fState |= MFS_DISABLED;
    SetMenuItemInfo( GUI.hMenu, ID_CHEAT_APPLY, FALSE, &mii);
    
    mii.fState = MFS_UNCHECKED;
    SetMenuItemInfo (GUI.hMenu, ID_SOUND_OPTIONS, FALSE, &mii);

    if (!Settings.APUEnabled)
        mii.fState |= MFS_DISABLED;
    
    SetMenuItemInfo (GUI.hMenu, ID_SOUND_NOSOUND, FALSE, &mii);
    for (i = 0; i < COUNT(SoundRates); i++)
        SetMenuItemInfo (GUI.hMenu, SoundRates[i].ident, FALSE, &mii);
    
    if (Settings.SoundPlaybackRate == 0)
        mii.fState |= MFS_DISABLED;

    SetMenuItemInfo (GUI.hMenu, ID_SOUND_16BIT, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_SOUND_STEREO, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_SOUND_SYNC, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_SOUND_INTERPOLATED, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_SOUND_25MS, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_SOUND_50MS, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_SOUND_100MS, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_SOUND_200MS, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_SOUND_500MS, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_SOUND_1S, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_SOUND_2S, FALSE, &mii);
    
    if (!Settings.Stereo)
        mii.fState |= MFS_DISABLED;

    SetMenuItemInfo (GUI.hMenu, ID_SOUND_REVERSE_STEREO, FALSE, &mii);

    mii.fState = !Settings.APUEnabled ? MFS_CHECKED | MFS_DISABLED : MFS_CHECKED;
    
    if (!DirectX.DSAvailable || Settings.SoundPlaybackRate == 0 || !Settings.APUEnabled)
        SetMenuItemInfo (GUI.hMenu, ID_SOUND_NOSOUND, FALSE, &mii);
    else
    {
        for (i = 0; i < COUNT(SoundRates); i++)
        {
            if (SoundRates [i].rate == Settings.SoundPlaybackRate)
            {
                SetMenuItemInfo (GUI.hMenu, SoundRates[i].ident, FALSE, &mii);
                break;
            }
        }
    }
    if (Settings.SoundPlaybackRate == 0)
        mii.fState |= MFS_DISABLED;

    int id;
    switch (Settings.SoundBufferSize)
    {
        case 1:  id = ID_SOUND_25MS; break;
        case 2:  id = ID_SOUND_50MS; break;
        default:
        case 4:  id = ID_SOUND_100MS; break;
        case 8:  id = ID_SOUND_200MS; break;
        case 16: id = ID_SOUND_500MS; break;
        case 32: id = ID_SOUND_1S; break;
        case 64: id = ID_SOUND_2S; break;
    }
    SetMenuItemInfo (GUI.hMenu, id, FALSE, &mii);
    
    if (Settings.SixteenBitSound)
        SetMenuItemInfo (GUI.hMenu, ID_SOUND_16BIT, FALSE, &mii);
    if (Settings.Stereo)
        SetMenuItemInfo (GUI.hMenu, ID_SOUND_STEREO, FALSE, &mii);
    if (Settings.InterpolatedSound)
        SetMenuItemInfo (GUI.hMenu, ID_SOUND_INTERPOLATED, FALSE,&mii);
    if (Settings.SoundSync)
        SetMenuItemInfo (GUI.hMenu, ID_SOUND_SYNC, FALSE, &mii);

    if (!Settings.Stereo)
        mii.fState |= MFS_DISABLED;
    if (Settings.ReverseStereo)
        SetMenuItemInfo (GUI.hMenu, ID_SOUND_REVERSE_STEREO, FALSE, &mii);

#ifndef DEBUGGER
    mii.fState = MFS_DISABLED;
#else
    mii.fState = MFS_UNCHECKED;
#endif
    SetMenuItemInfo (GUI.hMenu, ID_DEBUG_TRACE, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_DEBUG_TRACE_SPC, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_DEBUG_TRACE_SA1, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_DEBUG_TRACE_DSP1, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_DEBUG_FRAME_ADVANCE, FALSE, &mii);
    SetMenuItemInfo (GUI.hMenu, ID_DEBUG_SNES_STATUS, FALSE, &mii);
}

static void ResetFrameTimer ()
{
    if (GUI.hFrameTimer)
        timeKillEvent (GUI.hFrameTimer);

    GUI.hFrameTimer = timeSetEvent (Settings.FrameTime, 0, FrameTimer,
                                    0, TIME_PERIODIC);
}

bool8 LoadROM (const char *filename)
{
    if (Memory.LoadROM (filename))
    {
        SetupSound (Settings.SoundPlaybackRate, 
                    Settings.SixteenBitSound, 
                    Settings.Stereo);
        ResetFrameTimer ();
        return (TRUE);
    }
    return (FALSE);
}

bool8 S9xLoadROMImage (const char *string)
{
    RestoreGUIDisplay ();
    char *buf = new char [200 + strlen (string)];
    sprintf (buf, "\
The NetPlay server is requesting you load the following game:\n '%s'",
             string);

    MessageBox (GUI.hWnd, buf,
                "Snes9X: Information",
                MB_OK | MB_ICONINFORMATION);
    
    delete buf;

    char FileName [_MAX_PATH];
    if (_OpenROM (FileName, &Settings))
    {
        if (!Settings.StopEmulation)
        {
            Memory.SaveSRAM (S9xGetFilename (".srm"));
            S9xSaveCheatFile (S9xGetFilename (".cht"));
        }
        Settings.StopEmulation = !LoadROM (FileName);
        if (!Settings.StopEmulation) 
        {
            Memory.LoadSRAM( S9xGetFilename (".srm"));
            S9xLoadCheatFile (S9xGetFilename (".cht"));
            CheckDirectoryIsWritable (S9xGetFilename (".---"));
            CheckMenuStates ();
        }
        else
            return (FALSE);
    }
    else
        return (FALSE);

    return (TRUE);
}

/*****************************************************************************/

void EnableServer (bool8 enable)
{
    if (enable != Settings.NetPlayServer)
    {
        if (Settings.NetPlay && !Settings.NetPlayServer)
        {
            Settings.NetPlay = FALSE;
            S9xNPDisconnect ();
        }
                        
        if (enable)
        {
            S9xSetPause (PAUSE_NETPLAY_CONNECT);
            Settings.NetPlayServer = S9xNPStartServer (Settings.Port);
            Sleep (1000);
            
            if (!S9xNPConnectToServer ("127.0.0.1", Settings.Port,
                                       Memory.ROMName))
            {
                S9xClearPause (PAUSE_NETPLAY_CONNECT);
            }
        }
        else
        {
            Settings.NetPlayServer = FALSE;
            S9xNPStopServer ();
        }
    }
}

#ifdef USE_OPENGL
bool8 S9xOpenGLInit ()
{
    int PixelFormat;

    if (GUI.FullScreen)
    {
        DEVMODE dmScreenSettings;

        memset (&dmScreenSettings, 0, sizeof(dmScreenSettings));
        dmScreenSettings.dmSize = sizeof(dmScreenSettings);
        dmScreenSettings.dmPelsWidth = GUI.Width;
        dmScreenSettings.dmPelsHeight = GUI.Height;
        dmScreenSettings.dmBitsPerPel = 16; //bits;
        dmScreenSettings.dmFields = DM_BITSPERPEL|DM_PELSWIDTH|DM_PELSHEIGHT;

        if (ChangeDisplaySettings (&dmScreenSettings, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
        {
            GUI.FullScreen = FALSE;
        }
    }
    if (GUI.FullScreen)
    {
        SetWindowLong (GUI.hWnd, GWL_STYLE, WS_POPUP | WS_VISIBLE);
        SetWindowPos (GUI.hWnd, HWND_TOP, 0, 0, GUI.Width,
                      GUI.Height,
                      SWP_DRAWFRAME|SWP_FRAMECHANGED);
    }
    else
    {
        SetWindowLong (GUI.hWnd, GWL_STYLE, WS_POPUPWINDOW|WS_CAPTION|
                       WS_THICKFRAME|WS_VISIBLE|WS_MINIMIZEBOX|WS_MAXIMIZEBOX);
        SetWindowPos (GUI.hWnd, HWND_TOP, 
                      GUI.window_size.left,
                      GUI.window_size.top,
                      GUI.window_size.right - GUI.window_size.left,
                      GUI.window_size.bottom - GUI.window_size.top,
                      SWP_DRAWFRAME|SWP_FRAMECHANGED);
    }
    
    static PIXELFORMATDESCRIPTOR pfd =
    {
        sizeof(PIXELFORMATDESCRIPTOR),
        1,
        PFD_DRAW_TO_WINDOW |
        PFD_SUPPORT_OPENGL |
        PFD_DOUBLEBUFFER,
        PFD_TYPE_RGBA,
        16, //bits,
        0, 0, 0, 0, 0, 0,
        0,
        0,
        0,
        0, 0, 0, 0,
        16,
        0,
        0,
        PFD_MAIN_PLANE,
        0,
        0, 0, 0
    };

    if (!(PixelFormat = ChoosePixelFormat (GUI.hDC, &pfd)))
    {
        MessageBox(NULL,"Can't Find A Suitable PixelFormat.","ERROR",MB_OK|MB_ICONEXCLAMATION);
        return FALSE;
    }
    
    if (!SetPixelFormat (GUI.hDC, PixelFormat, &pfd))
    {
        MessageBox(NULL,"Can't Set The PixelFormat.","ERROR",MB_OK|MB_ICONEXCLAMATION);
        return (FALSE);
    }
    
    if (!(GUI.hRC = wglCreateContext (GUI.hDC)))
    {
        MessageBox(NULL,"Can't Create A GL Rendering Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
        return (FALSE);
    }
    
    if (!wglMakeCurrent (GUI.hDC, GUI.hRC))
    {
        MessageBox(NULL,"Can't Activate The GL Rendering Context.","ERROR",MB_OK|MB_ICONEXCLAMATION);
        return (FALSE);
    }
    
    glGetIntegerv (GL_MAX_TEXTURE_SIZE, &OpenGL.max_texture_size);

    if (OpenGL.max_texture_size >= 512)
    {
        OpenGL.texture_size = 512;
        OpenGL.num_textures = 2;
    }
    else
    {
        OpenGL.texture_size = OpenGL.max_texture_size;
        OpenGL.num_textures = 1;
    }

    Settings.OpenGLEnable = TRUE;

    const char *ext = (const char *) glGetString (GL_EXTENSIONS);

    if (ext && strstr (ext, "EXT_packed_pixels") != NULL)
        OpenGL.packed_pixels_extension_present = TRUE;

    const char *version = (const char *) glGetString (GL_VERSION);

    if (version && strlen (version) < 100)
    {
	char ver [100];
	strcpy (ver,  version);

	// Strip dots from the version string
	char *ptr;
	while (ptr = strchr (ver, '.'))
	    memmove (ptr, ptr + 1, strlen (ptr + 1) + 1);
	
	// Pad the string with zeros to 4 digits
	while (strlen (ver) < 4)
	    strcat (ver, "0");

	OpenGL.version = atoi (ver);
    }
    else
	OpenGL.version = 1100;

#ifndef GL_UNSIGNED_SHORT_5_5_5_1_EXT
#define GL_UNSIGNED_SHORT_5_5_5_1_EXT     0x8034
#endif

    if (OpenGL.version >= 1200)
    {
        OpenGL.internal_format = GL_RGB5_A1;
        OpenGL.format = GL_RGBA;
        OpenGL.type = GL_UNSIGNED_SHORT_5_5_5_1_EXT;
    }
    else
    if (OpenGL.packed_pixels_extension_present)
    {
        OpenGL.internal_format = GL_RGB5_A1;
        OpenGL.format = GL_RGBA;
        OpenGL.type = GL_UNSIGNED_SHORT_5_5_5_1_EXT;
    }
    else
    {
        OpenGL.internal_format = GL_RGB;
        OpenGL.format = GL_RGB;
        OpenGL.type = GL_UNSIGNED_BYTE;
    }

    glGenTextures (OpenGL.num_textures, OpenGL.textures);

    if (OpenGL.num_textures == 2)
    {
        glBindTexture (GL_TEXTURE_2D, OpenGL.textures [1]);
        glTexImage2D (GL_TEXTURE_2D, 0, OpenGL.internal_format, 256, 256, 0,
                      OpenGL.format, OpenGL.type, NULL);

        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    }

    glBindTexture (GL_TEXTURE_2D, OpenGL.textures [0]);
    glTexImage2D (GL_TEXTURE_2D, 0, OpenGL.internal_format, 
                  OpenGL.texture_size, OpenGL.texture_size, 0, 
                  OpenGL.format, OpenGL.type, NULL);

    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    glPolygonMode (GL_FRONT, GL_FILL);
    glEnable (GL_CULL_FACE);
    glCullFace (GL_BACK);

    glEnable (GL_DITHER);
    glEnable (GL_LIGHTING);
    glEnable (GL_LIGHT0);

    glEnable (GL_POINT_SMOOTH);
    glHint (GL_POINT_SMOOTH_HINT, GL_NICEST);

    glEnable (GL_TEXTURE_2D);

    glShadeModel (GL_SMOOTH);
    glClearColor (0.0f, 0.0f, 0.0f, 0.0f);
    glClearDepth (1.0f);
    glDisable (GL_DEPTH_TEST);
    glDepthFunc (GL_LESS);
    glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

    GLfloat LightAmbient[]    = { 0.2f, 0.2f, 0.2f, 1.0f };
    GLfloat LightDiffuse[]    = { 1.0f, 1.0f, 1.0f, 1.0f };
    GLfloat LightSpecular[]   = { 0.5f, 0.5f, 0.5f, 1.0f };
    GLfloat LightPosition[]   = { 0.0f, 0.0f, 2.0f, 1.0f };

    glLightfv (GL_LIGHT0, GL_AMBIENT, LightAmbient);
    glLightfv (GL_LIGHT0, GL_DIFFUSE, LightDiffuse);
    glLightfv (GL_LIGHT0, GL_SPECULAR, LightSpecular);
    glLightfv (GL_LIGHT0, GL_POSITION, LightPosition);

    //Set common material properties
    GLfloat MatSpecular[]    = { 1.0f, 1.0f, 1.0f, 1.1f };
    GLfloat WhMat[]          = { 1.0f, 1.0f, 1.0f, 1.0f };
    
    glMaterialf (GL_FRONT_AND_BACK, GL_SHININESS, 128.0f);
    glMaterialfv (GL_FRONT_AND_BACK, GL_SPECULAR, MatSpecular);
    glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, WhMat);
    
    glMatrixMode (GL_PROJECTION);
    glLoadIdentity ();
    glMatrixMode (GL_MODELVIEW);

    return (TRUE);
}

void S9xOpenGLDeinit ()
{
    if (GUI.FullScreen)
        ChangeDisplaySettings (NULL, 0);

    if (GUI.hRC)
    {
        wglMakeCurrent (NULL, NULL);
        wglDeleteContext (GUI.hRC);
        GUI.hRC = NULL;
    }
}

void S9xOpenGLResize (int width, int height)
{
    if (height == 0)
        height = 1;

    glViewport (0, 0, width, height);

    glMatrixMode (GL_PROJECTION);
    glLoadIdentity();

    gluPerspective (45.0f, (GLfloat)width / (GLfloat)height, 0.1f, 100.0f);
    glMatrixMode (GL_MODELVIEW);
}
#endif

void S9xAddToRecentGames (const char *filename)
{
    if (!GUI.RecentGames)
    {
        GUI.RecentGames = new char *[2];
        GUI.RecentGames [0] = strdup (filename);
        GUI.RecentGames [1] = NULL;
    }
    else
    {
        int i = -1;

        // Make sure its not in the list already
        while (GUI.RecentGames [++i])
            if (strcmp (filename, GUI.RecentGames [i]) == 0)
                break;

        if (GUI.RecentGames [i])
        {
            // It is in the list, move it to the head of the list.
            char *t = GUI.RecentGames [i];
            for (int j = i; j > 0; j--)
                GUI.RecentGames [j] = GUI.RecentGames [j - 1];

            GUI.RecentGames [0] = t;
        }
        else
        {
            // Not in the list, add it.
            if (i < MAX_RECENT_GAMES_LIST_SIZE)
            {
                // Extend the recent game list length by 1.
                char **t = new char *[i + 2];

                memmove (&t [1], GUI.RecentGames, sizeof (char *) * (i + 1));
                delete GUI.RecentGames;
                GUI.RecentGames = t;
                GUI.RecentGames [0] = strdup (filename);
            }
            else
            {
                // Throw the last item off the end of the list
                free (GUI.RecentGames [i - 1]);
                memmove (&GUI.RecentGames [1], &GUI.RecentGames [0],
                         sizeof (char *) * (MAX_RECENT_GAMES_LIST_SIZE - 1));
                GUI.RecentGames [0] = strdup (filename);
            }
        }
    }
    S9xSetRecentGames ();
}

void S9xSetRecentGames ()
{
    HMENU file = GetSubMenu (GUI.hMenu, 0);
    if (file)
    {
        HMENU recent = GetSubMenu (file, 1);
        if (recent)
        {
            MENUITEMINFO mii;
            char name [256 + 10];
            int i;

            // Clear out the menu first
            for (i = GetMenuItemCount (recent) - 1; i >= 0; i--)
                RemoveMenu (recent, i, MF_BYPOSITION);

            mii.cbSize = sizeof (mii);
            mii.fMask = MIIM_TYPE | MIIM_DATA | MIIM_STATE | MIIM_ID;
            mii.fType = MFT_STRING;
            mii.fState = MFS_UNCHECKED;

            for (i = 0; i < 10 && GUI.RecentGames && GUI.RecentGames [i]; i++)
            {
                // Build up a menu item string in the form:
                // 1. <basename of ROM image name>
                sprintf (name, "&%c. %s", i < 9 ? '1' + i : 'A' + i - 9,
                         S9xBasename (GUI.RecentGames [i]));
                mii.dwTypeData = name;
                mii.cch = strlen (name) + 1;
                mii.wID = 0xFF00 + i;

                InsertMenuItem (recent, 0xFF00 + i, FALSE, &mii);
            }
        }
    }
}

static void WinDeleteRecentGamesList ()
{
    if (GUI.RecentGames)
    {
        int i = 0;

        while (GUI.RecentGames [i])
            free (GUI.RecentGames [i++]);

        delete GUI.RecentGames;
        GUI.RecentGames = NULL;
    }
}

int CALLBACK SoundConf(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static struct SSettings* set;
	switch(msg)
	{
		case WM_INITDIALOG:

			set= (struct SSettings *)lParam;

			SendDlgItemMessage(hDlg, IDC_DRIVER, CB_INSERTSTRING,0,(LPARAM)TEXT("Snes9x DirectSound"));
			SendDlgItemMessage(hDlg, IDC_DRIVER, CB_INSERTSTRING,1,(LPARAM)TEXT("FMOD DirectSound"));
			SendDlgItemMessage(hDlg, IDC_DRIVER, CB_INSERTSTRING,2,(LPARAM)TEXT("FMOD Windows Multimedia"));
			SendDlgItemMessage(hDlg, IDC_DRIVER, CB_INSERTSTRING,3,(LPARAM)TEXT("FMOD A3D"));
			SendDlgItemMessage(hDlg, IDC_DRIVER,CB_SETCURSEL,set->SoundDriver,0);

			SendDlgItemMessage(hDlg, IDC_SKIP_TYPE, CB_INSERTSTRING,0,(LPARAM)TEXT("Skip style #1"));
			SendDlgItemMessage(hDlg, IDC_SKIP_TYPE, CB_INSERTSTRING,1,(LPARAM)TEXT("Skip style #2"));
			SendDlgItemMessage(hDlg, IDC_SKIP_TYPE, CB_INSERTSTRING,2,(LPARAM)TEXT("Skip style #3"));
			SendDlgItemMessage(hDlg, IDC_SKIP_TYPE, CB_INSERTSTRING,3,(LPARAM)TEXT("Skip style #4"));
			SendDlgItemMessage(hDlg, IDC_SKIP_TYPE, CB_INSERTSTRING,4,(LPARAM)TEXT("Skip style #5"));

			SendDlgItemMessage(hDlg,IDC_SKIP_TYPE,CB_SETCURSEL,set->SoundSkipMethod,0);

			SendDlgItemMessage(hDlg, IDC_RATE, CB_INSERTSTRING,0,(LPARAM)TEXT("<No Sound>"));
			SendDlgItemMessage(hDlg, IDC_RATE, CB_INSERTSTRING,1,(LPARAM)TEXT("8 KHz"));
			SendDlgItemMessage(hDlg, IDC_RATE, CB_INSERTSTRING,2,(LPARAM)TEXT("11 KHz"));
			SendDlgItemMessage(hDlg, IDC_RATE, CB_INSERTSTRING,3,(LPARAM)TEXT("16 KHz"));
			SendDlgItemMessage(hDlg, IDC_RATE, CB_INSERTSTRING,4,(LPARAM)TEXT("22 KHz"));
			SendDlgItemMessage(hDlg, IDC_RATE, CB_INSERTSTRING,5,(LPARAM)TEXT("30 KHz"));
			SendDlgItemMessage(hDlg, IDC_RATE, CB_INSERTSTRING,6,(LPARAM)TEXT("32 KHz"));
			SendDlgItemMessage(hDlg, IDC_RATE, CB_INSERTSTRING,7,(LPARAM)TEXT("35 KHz"));
			SendDlgItemMessage(hDlg, IDC_RATE, CB_INSERTSTRING,8,(LPARAM)TEXT("44 KHz"));
			SendDlgItemMessage(hDlg, IDC_RATE, CB_INSERTSTRING,9,(LPARAM)TEXT("48 KHz"));

			int temp;
			switch(set->SoundPlaybackRate)
			{
			case 8000:temp=1;break;
			case 11000:temp=2;break;
			case 16000:temp=3;break;
			case 22000:temp=4;break;
			case 30000:temp=5;break;
			case 32000:temp=6;break;
			case 35000:temp=7;break;
			case 44000:
			case 44100:temp=8;break;
			case 48000:temp=9;break;
			case 0:
			default:temp=0;break;
			}
			SendDlgItemMessage(hDlg,IDC_RATE,CB_SETCURSEL,temp,0);

			SendDlgItemMessage(hDlg, IDC_MIX, CB_INSERTSTRING,0,(LPARAM)TEXT("10 ms"));
			SendDlgItemMessage(hDlg, IDC_MIX, CB_INSERTSTRING,1,(LPARAM)TEXT("20 ms"));
			SendDlgItemMessage(hDlg, IDC_MIX, CB_INSERTSTRING,2,(LPARAM)TEXT("30 ms"));
			SendDlgItemMessage(hDlg, IDC_MIX, CB_INSERTSTRING,3,(LPARAM)TEXT("40 ms"));
			SendDlgItemMessage(hDlg, IDC_MIX, CB_INSERTSTRING,4,(LPARAM)TEXT("50 ms"));
			SendDlgItemMessage(hDlg, IDC_MIX, CB_INSERTSTRING,5,(LPARAM)TEXT("60 ms"));
			SendDlgItemMessage(hDlg, IDC_MIX, CB_INSERTSTRING,6,(LPARAM)TEXT("70 ms"));
			SendDlgItemMessage(hDlg, IDC_MIX, CB_INSERTSTRING,7,(LPARAM)TEXT("80 ms"));
			SendDlgItemMessage(hDlg, IDC_MIX, CB_INSERTSTRING,8,(LPARAM)TEXT("90 ms"));
			SendDlgItemMessage(hDlg, IDC_MIX, CB_INSERTSTRING,9,(LPARAM)TEXT("100 ms"));
			SendDlgItemMessage(hDlg, IDC_MIX, CB_INSERTSTRING,10,(LPARAM)TEXT("110 ms"));
			SendDlgItemMessage(hDlg, IDC_MIX, CB_INSERTSTRING,11,(LPARAM)TEXT("120 ms"));
			SendDlgItemMessage(hDlg, IDC_MIX, CB_INSERTSTRING,12,(LPARAM)TEXT("130 ms"));
			SendDlgItemMessage(hDlg, IDC_MIX, CB_INSERTSTRING,13,(LPARAM)TEXT("140 ms"));
			SendDlgItemMessage(hDlg, IDC_MIX, CB_INSERTSTRING,14,(LPARAM)TEXT("150 ms"));

			SendDlgItemMessage(hDlg,IDC_MIX,CB_SETCURSEL,((set->SoundMixInterval/10)-1),0);

			SendDlgItemMessage(hDlg, IDC_BUFLEN, CB_INSERTSTRING,0,(LPARAM)TEXT("10 ms"));
			SendDlgItemMessage(hDlg, IDC_BUFLEN, CB_INSERTSTRING,1,(LPARAM)TEXT("20 ms"));
			SendDlgItemMessage(hDlg, IDC_BUFLEN, CB_INSERTSTRING,2,(LPARAM)TEXT("40 ms"));
			SendDlgItemMessage(hDlg, IDC_BUFLEN, CB_INSERTSTRING,3,(LPARAM)TEXT("80 ms"));
			SendDlgItemMessage(hDlg, IDC_BUFLEN, CB_INSERTSTRING,4,(LPARAM)TEXT("160 ms"));
			SendDlgItemMessage(hDlg, IDC_BUFLEN, CB_INSERTSTRING,5,(LPARAM)TEXT("320 ms"));
			SendDlgItemMessage(hDlg, IDC_BUFLEN, CB_INSERTSTRING,6,(LPARAM)TEXT("640 ms"));
			
			switch(set->SoundBufferSize)
			{
			case 2:temp=1; break;
			case 4:temp=2; break;
			case 8:temp=3; break;
			case 16:temp=4; break;
			case 32:temp=5; break;
			case 64: temp=6;break;
			case 1:
			default:temp=0;break;
			}
			SendDlgItemMessage(hDlg,IDC_BUFLEN,CB_SETCURSEL,temp,0);

			if(set->SixteenBitSound)
				SendDlgItemMessage(hDlg,IDC_16BIT,BM_SETCHECK,BST_CHECKED,0);
			if(set->Stereo)
				SendDlgItemMessage(hDlg,IDC_STEREO,BM_SETCHECK,BST_CHECKED,0);
			else EnableWindow(GetDlgItem(hDlg, IDC_REV_STEREO), FALSE);
			if(set->ReverseStereo)
				SendDlgItemMessage(hDlg,IDC_REV_STEREO,BM_SETCHECK,BST_CHECKED,0);
			if(set->AltSampleDecode)
			{
				EnableWindow(GetDlgItem(hDlg, IDC_CACHING), FALSE);
				SendDlgItemMessage(hDlg,IDC_ANTIRES,BM_SETCHECK,BST_CHECKED,0);
			}
			if(set->Mute)
				SendDlgItemMessage(hDlg,IDC_MUTE,BM_SETCHECK,BST_CHECKED,0);

			if(set->NextAPUEnabled)
			{
				EnableWindow(GetDlgItem(hDlg, IDC_SKIP_TYPE), FALSE);
				SendDlgItemMessage(hDlg,IDC_SPC700ON,BM_SETCHECK,BST_CHECKED,0);
				//Gray out the skip feature!
			}
			if(set->InterpolatedSound)
				SendDlgItemMessage(hDlg,IDC_LINEAR_INTER,BM_SETCHECK,BST_CHECKED,0);
			if(set->SoundSync)
				SendDlgItemMessage(hDlg,IDC_SYNC_TO_SOUND_CPU,BM_SETCHECK,BST_CHECKED,0);
			if(set->SoundEnvelopeHeightReading)
				SendDlgItemMessage(hDlg,IDC_ENVX,BM_SETCHECK,BST_CHECKED,0);
			if(!set->DisableSoundEcho)
				SendDlgItemMessage(hDlg,IDC_ECHO,BM_SETCHECK,BST_CHECKED,0);

			if(!set->DisableSampleCaching)
				SendDlgItemMessage(hDlg,IDC_CACHING,BM_SETCHECK,BST_CHECKED,0);
			if(!set->DisableMasterVolume)
				SendDlgItemMessage(hDlg,IDC_MASTER_VOL,BM_SETCHECK,BST_CHECKED,0);

			if(set->SoundDriver)
			{
				EnableWindow(GetDlgItem(hDlg, IDC_MIX), false);
				EnableWindow(GetDlgItem(hDlg, IDC_BUFLEN), false);
				EnableWindow(GetDlgItem(hDlg, IDC_SYNC_TO_SOUND_CPU), false);
			}

			if(!IsDlgButtonChecked(hDlg,IDC_SPC700ON))
			{
				EnableWindow(GetDlgItem(hDlg, IDC_CACHING), FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_MUTE), FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_ANTIRES), FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_SKIP_TYPE), TRUE);
				EnableWindow(GetDlgItem(hDlg, IDC_DRIVER), FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_16BIT), FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_ENVX), FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_STEREO), FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_MASTER_VOL), FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_RATE), FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_ECHO), FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_LINEAR_INTER), FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_REV_STEREO), FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_BUFLEN), FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_MIX), FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_SYNC_TO_SOUND_CPU), FALSE);

			}

			return true;
		case WM_COMMAND:
		switch(LOWORD(wParam))
		{
			case IDOK:
				
			set->SoundDriver=SendDlgItemMessage(hDlg, IDC_DRIVER,CB_GETCURSEL,0,0);
			set->SoundSkipMethod=(unsigned char)SendDlgItemMessage(hDlg,IDC_SKIP_TYPE,CB_GETCURSEL,0,0);
			set->SixteenBitSound=IsDlgButtonChecked(hDlg, IDC_16BIT);
			set->AltSampleDecode=IsDlgButtonChecked(hDlg, IDC_ANTIRES);
			set->SoundSync=IsDlgButtonChecked(hDlg, IDC_SYNC_TO_SOUND_CPU);
			set->InterpolatedSound=IsDlgButtonChecked(hDlg, IDC_LINEAR_INTER);
			set->Stereo=IsDlgButtonChecked(hDlg, IDC_STEREO);
			set->ReverseStereo=IsDlgButtonChecked(hDlg, IDC_REV_STEREO);
			set->Mute=IsDlgButtonChecked(hDlg, IDC_MUTE);
			set->NextAPUEnabled=IsDlgButtonChecked(hDlg, IDC_SPC700ON);
			set->SoundEnvelopeHeightReading=IsDlgButtonChecked(hDlg, IDC_ENVX);
			
			set->DisableSoundEcho=(!IsDlgButtonChecked(hDlg, IDC_ECHO));
			set->DisableSampleCaching=(!IsDlgButtonChecked(hDlg, IDC_CACHING));
			set->DisableMasterVolume=(!IsDlgButtonChecked(hDlg, IDC_MASTER_VOL));


			switch(SendDlgItemMessage(hDlg, IDC_RATE,CB_GETCURSEL,0,0))
			{
			case 0: set->SoundPlaybackRate=0;		break;
			case 1: set->SoundPlaybackRate=8000;	break;
			case 2: set->SoundPlaybackRate=11000;	break;
			case 3: set->SoundPlaybackRate=16000;	break;
			case 4: set->SoundPlaybackRate=22000;	break;
			case 5: set->SoundPlaybackRate=30000;	break;
			case 6: set->SoundPlaybackRate=32000;	break;
			case 7: set->SoundPlaybackRate=35000;	break;
			case 8: set->SoundPlaybackRate=44100;	break;
			case 9: set->SoundPlaybackRate=48000;	break;
			}
			set->SoundMixInterval=(10*(1+(SendDlgItemMessage(hDlg,IDC_MIX,CB_GETCURSEL,0,0))));

			set->SoundBufferSize=1<<SendDlgItemMessage(hDlg,IDC_BUFLEN,CB_GETCURSEL,0,0);
			case IDCANCEL:
				EndDialog(hDlg, 1);
				return true;
			
			case IDC_SPC700ON:
				if(BN_CLICKED==HIWORD(wParam)||BN_DBLCLK==HIWORD(wParam))
				{
					if(!IsDlgButtonChecked(hDlg,IDC_SPC700ON))
					{
						EnableWindow(GetDlgItem(hDlg, IDC_CACHING), FALSE);
						EnableWindow(GetDlgItem(hDlg, IDC_MUTE), FALSE);
						EnableWindow(GetDlgItem(hDlg, IDC_ANTIRES), FALSE);
						EnableWindow(GetDlgItem(hDlg, IDC_SKIP_TYPE), TRUE);
						EnableWindow(GetDlgItem(hDlg, IDC_DRIVER), FALSE);
						EnableWindow(GetDlgItem(hDlg, IDC_16BIT), FALSE);
						EnableWindow(GetDlgItem(hDlg, IDC_ENVX), FALSE);
						EnableWindow(GetDlgItem(hDlg, IDC_STEREO), FALSE);
						EnableWindow(GetDlgItem(hDlg, IDC_MASTER_VOL), FALSE);
						EnableWindow(GetDlgItem(hDlg, IDC_RATE), FALSE);
						EnableWindow(GetDlgItem(hDlg, IDC_ECHO), FALSE);
						EnableWindow(GetDlgItem(hDlg, IDC_LINEAR_INTER), FALSE);
						EnableWindow(GetDlgItem(hDlg, IDC_REV_STEREO), FALSE);
						EnableWindow(GetDlgItem(hDlg, IDC_BUFLEN), FALSE);
						EnableWindow(GetDlgItem(hDlg, IDC_MIX), FALSE);
						EnableWindow(GetDlgItem(hDlg, IDC_SYNC_TO_SOUND_CPU), FALSE);

					}
					else
					{
						EnableWindow(GetDlgItem(hDlg, IDC_LINEAR_INTER), TRUE);
						EnableWindow(GetDlgItem(hDlg, IDC_ECHO), TRUE);
						EnableWindow(GetDlgItem(hDlg, IDC_RATE), TRUE);
						EnableWindow(GetDlgItem(hDlg, IDC_MASTER_VOL), TRUE);
						EnableWindow(GetDlgItem(hDlg, IDC_STEREO), TRUE);
						EnableWindow(GetDlgItem(hDlg, IDC_DRIVER), TRUE);
						EnableWindow(GetDlgItem(hDlg, IDC_16BIT), TRUE);
						EnableWindow(GetDlgItem(hDlg, IDC_ENVX), TRUE);
						EnableWindow(GetDlgItem(hDlg, IDC_SKIP_TYPE), FALSE);
						EnableWindow(GetDlgItem(hDlg, IDC_ANTIRES), TRUE);
						EnableWindow(GetDlgItem(hDlg, IDC_MUTE), TRUE);
						if(!IsDlgButtonChecked(hDlg,IDC_ANTIRES))
							EnableWindow(GetDlgItem(hDlg, IDC_CACHING), TRUE);
						if(IsDlgButtonChecked(hDlg,IDC_STEREO))
							EnableWindow(GetDlgItem(hDlg, IDC_REV_STEREO), TRUE);
						if(0==SendDlgItemMessage(hDlg, IDC_DRIVER, CB_GETCURSEL, 0,0))
						{
							//enable stuff.
							EnableWindow(GetDlgItem(hDlg, IDC_MIX), true);
							EnableWindow(GetDlgItem(hDlg, IDC_BUFLEN), true);
							EnableWindow(GetDlgItem(hDlg, IDC_SYNC_TO_SOUND_CPU), true);
						
						}
					return true;
						//Check Driver States for other options.
					}
					return true;
				}
				else return false;
			case IDC_DRIVER:
				if(CBN_SELCHANGE==HIWORD(wParam))
				{
					//get index
					int i;
					i=SendDlgItemMessage(hDlg, IDC_DRIVER, CB_GETCURSEL, 0,0);
					if(i)
					{
						EnableWindow(GetDlgItem(hDlg, IDC_MIX), false);
						EnableWindow(GetDlgItem(hDlg, IDC_BUFLEN), false);
						EnableWindow(GetDlgItem(hDlg, IDC_SYNC_TO_SOUND_CPU), false);
					}
					//disable if index...
					else
					{
						//enable stuff.
						EnableWindow(GetDlgItem(hDlg, IDC_MIX), true);
						EnableWindow(GetDlgItem(hDlg, IDC_BUFLEN), true);
						EnableWindow(GetDlgItem(hDlg, IDC_SYNC_TO_SOUND_CPU), true);
					
					}
					return true;
				}
				else return false;
			case IDC_ANTIRES:
			{
				if(BN_CLICKED==HIWORD(wParam)||BN_DBLCLK==HIWORD(wParam))
					{
						if(!IsDlgButtonChecked(hDlg,IDC_ANTIRES))
						{
							EnableWindow(GetDlgItem(hDlg, IDC_CACHING), TRUE);
						}
						else EnableWindow(GetDlgItem(hDlg, IDC_CACHING), FALSE);
						return true;
 
					}
					else return false;
			}
			case IDC_STEREO:
			{
				if(BN_CLICKED==HIWORD(wParam)||BN_DBLCLK==HIWORD(wParam))
					{
						if(IsDlgButtonChecked(hDlg,IDC_STEREO))
						{
							EnableWindow(GetDlgItem(hDlg, IDC_REV_STEREO), TRUE);
						}
						else EnableWindow(GetDlgItem(hDlg, IDC_REV_STEREO), FALSE);
						return true;
 
					}
					else return false;
			}
			default: return false;
		

		}
	}
	return false;
}

int CALLBACK SP7PackConfig(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
	case WM_INITDIALOG:
		{
			if(LoadUp7110==&SPC7110Grab)
				SendDlgItemMessage(hDlg, IDC_SPC7110_SOME, BM_SETCHECK, (WPARAM)BST_CHECKED, 0);
			else if(LoadUp7110==&SPC7110Open)
				SendDlgItemMessage(hDlg, IDC_SPC7110_FILE, BM_SETCHECK, (WPARAM)BST_CHECKED, 0);
			else	SendDlgItemMessage(hDlg, IDC_SPC7110_ALL, BM_SETCHECK, (WPARAM)BST_CHECKED, 0);
			SendDlgItemMessage(hDlg, IDC_SPIN_CACHE, UDM_SETRANGE, 0, MAKELONG(20,1));
			SendDlgItemMessage(hDlg, IDC_SPIN_CACHE, UDM_SETPOS, 0, MAKELONG(cacheMegs,0));
			return true;
		}
	case WM_COMMAND:
		{
			switch(wParam)
			{
			case IDOK:
				if(BST_CHECKED==SendDlgItemMessage(hDlg, IDC_SPC7110_ALL, BM_GETCHECK,0,0))
				{
					LoadUp7110=&SPC7110Load;
				}
				else if(BST_CHECKED==SendDlgItemMessage(hDlg, IDC_SPC7110_FILE, BM_GETCHECK,0,0))
				{
					LoadUp7110=&SPC7110Open;
				}
				else
				{
					LoadUp7110=&SPC7110Grab;
					cacheMegs=(uint16)SendDlgItemMessage(hDlg,IDC_SPIN_CACHE,UDM_GETPOS,0,0);
				}
			case IDCANCEL:
				EndDialog(hDlg,0);
			default: return false;
			}
		}
	default: return false;
	}
}

#ifdef RTC_DEBUGGER
int CALLBACK SPC7110rtc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static int month;
	static int day;
	static int year;
	static int hour;
	static int minutes;
	static int seconds;
	static int dayinmonth;
	static struct SPC7110RTC* rtc;
	switch(msg)
	{
		case WM_INITDIALOG:

			rtc= (struct SPC7110RTC *)lParam;

			seconds=rtc->reg[0]+rtc->reg[1]*10;
			minutes=rtc->reg[2]+rtc->reg[3]*10;
			hour=rtc->reg[4]+rtc->reg[5]*10;
			day=rtc->reg[6]+rtc->reg[7]*10;
			month=rtc->reg[8]+rtc->reg[9]*10;
			year=rtc->reg[10]+rtc->reg[11]*10;
			dayinmonth=S9xRTCDaysInMonth(month,year);


			SendDlgItemMessage(hDlg,IDC_MONTH, TBM_SETRANGE,(WPARAM)1,(LPARAM)MAKELONG(0,11));
			SendDlgItemMessage(hDlg,IDC_MONTH, TBM_SETTICFREQ,(WPARAM)5,0);
			SendDlgItemMessage(hDlg,IDC_MONTH, TBM_SETPOS,(WPARAM)(BOOL)TRUE,(LPARAM)month-1);
			SetDlgItemInt(hDlg,IDC_LBLMONTH,month,0);
			
			SendDlgItemMessage(hDlg,IDC_DAY, TBM_SETRANGE,(WPARAM)1,(LPARAM)MAKELONG(0,dayinmonth-1));
			SendDlgItemMessage(hDlg,IDC_DAY, TBM_SETTICFREQ,(WPARAM)5,0);
			SendDlgItemMessage(hDlg,IDC_DAY, TBM_SETPOS,(WPARAM)(BOOL)TRUE,(LPARAM)day-1);
			SetDlgItemInt(hDlg,IDC_LBLDAY,day,0);

			SendDlgItemMessage(hDlg,IDC_YEAR, TBM_SETRANGE,(WPARAM)1,(LPARAM)MAKELONG(0,99));
			SendDlgItemMessage(hDlg,IDC_YEAR, TBM_SETTICFREQ,(WPARAM)5,0);
			if(year<95)
			{
				SendDlgItemMessage(hDlg,IDC_YEAR, TBM_SETPOS,(WPARAM)(BOOL)TRUE,(LPARAM)year+5);
				SetDlgItemInt(hDlg,IDC_LBLYEAR,year+2000,0);
			}
			else
			{
				SendDlgItemMessage(hDlg,IDC_YEAR, TBM_SETPOS,(WPARAM)(BOOL)TRUE,(LPARAM)year-95);
				SetDlgItemInt(hDlg,IDC_LBLYEAR,year+1900,0);
			}

			SendDlgItemMessage(hDlg, IDC_HOUR, TBM_SETRANGE,(WPARAM)1,(LPARAM)MAKELONG(0,23));
			SendDlgItemMessage(hDlg,IDC_HOUR, TBM_SETTICFREQ,(WPARAM)5,0);
			SendDlgItemMessage(hDlg,IDC_HOUR, TBM_SETPOS,(WPARAM)(BOOL)TRUE,(LPARAM)hour);
			SetDlgItemInt(hDlg,IDC_LBLHOUR,hour,0);

			SendDlgItemMessage(hDlg, IDC_MINUTE, TBM_SETRANGE,(WPARAM)1,(LPARAM)MAKELONG(0,59));
			SendDlgItemMessage(hDlg,IDC_MINUTE, TBM_SETTICFREQ,(WPARAM)5,0);
			SendDlgItemMessage(hDlg,IDC_MINUTE, TBM_SETPOS,(WPARAM)(BOOL)TRUE,(LPARAM)minutes);
			SetDlgItemInt(hDlg,IDC_LBLMINUTE,minutes,0);

			SendDlgItemMessage(hDlg, IDC_SECOND, TBM_SETRANGE,(WPARAM)1,(LPARAM)MAKELONG(0,59));
			SendDlgItemMessage(hDlg,IDC_SECOND, TBM_SETTICFREQ,(WPARAM)5,0);
			SendDlgItemMessage(hDlg,IDC_SECOND, TBM_SETPOS,(WPARAM)(BOOL)TRUE,(LPARAM)seconds);
			SetDlgItemInt(hDlg,IDC_LBLSECOND,seconds,0);



			if(rtc->reg[0x0D]&0x01)
				SendDlgItemMessage(hDlg,IDC_RTC_D1,BM_SETCHECK,BST_CHECKED,0);
			if(rtc->reg[0x0D]&0x02)
				SendDlgItemMessage(hDlg,IDC_RTC_D2,BM_SETCHECK,BST_CHECKED,0);
			if(rtc->reg[0x0D]&0x04)
				SendDlgItemMessage(hDlg,IDC_RTC_D4,BM_SETCHECK,BST_CHECKED,0);
			if(rtc->reg[0x0D]&0x08)
				SendDlgItemMessage(hDlg,IDC_RTC_D8,BM_SETCHECK,BST_CHECKED,0);
			
			if(rtc->reg[0x0E]&0x01)
				SendDlgItemMessage(hDlg,IDC_RTC_E1,BM_SETCHECK,BST_CHECKED,0);
			if(rtc->reg[0x0E]&0x02)
				SendDlgItemMessage(hDlg,IDC_RTC_E2,BM_SETCHECK,BST_CHECKED,0);
			if(rtc->reg[0x0E]&0x04)
				SendDlgItemMessage(hDlg,IDC_RTC_E4,BM_SETCHECK,BST_CHECKED,0);
			if(rtc->reg[0x0E]&0x08)
				SendDlgItemMessage(hDlg,IDC_RTC_E8,BM_SETCHECK,BST_CHECKED,0);
			
			if(rtc->reg[0x0F]&0x01)
				SendDlgItemMessage(hDlg,IDC_RTC_F1,BM_SETCHECK,BST_CHECKED,0);
			if(rtc->reg[0x0F]&0x02)
				SendDlgItemMessage(hDlg,IDC_RTC_F2,BM_SETCHECK,BST_CHECKED,0);
			if(rtc->reg[0x0F]&0x04)
				SendDlgItemMessage(hDlg,IDC_RTC_F4,BM_SETCHECK,BST_CHECKED,0);
			if(rtc->reg[0x0F]&0x08)
				SendDlgItemMessage(hDlg,IDC_RTC_F8,BM_SETCHECK,BST_CHECKED,0);
			
			return true;
		case  WM_COMMAND:
			switch(LOWORD(wParam))
			{
				case IDOK:
					rtc->reg[0x0]=seconds%10;
					rtc->reg[0x1]=seconds/10;
					rtc->reg[0x2]=minutes%10;
					rtc->reg[0x3]=minutes/10;
					rtc->reg[0x4]=hour%10;
					rtc->reg[0x5]=hour/10;
					rtc->reg[0x6]=day%10;
					rtc->reg[0x7]=day/10;
					rtc->reg[0x8]=month%10;
					rtc->reg[0x9]=month/10;
					rtc->reg[0xA]=year%10;
					rtc->reg[0xB]=year/10;

					if(IsDlgButtonChecked(hDlg, IDC_RTC_D1))
						rtc->reg[0x0D]|=0x01;
					else
						rtc->reg[0x0D]&=0x0E;
					if(IsDlgButtonChecked(hDlg, IDC_RTC_D2))
						rtc->reg[0x0D]|=0x02;
					else
						rtc->reg[0x0D]&=0x0D;
					if(IsDlgButtonChecked(hDlg, IDC_RTC_D4))
						rtc->reg[0x0D]|=0x04;
					else
						rtc->reg[0x0D]&=0x0B;
					if(IsDlgButtonChecked(hDlg, IDC_RTC_D8))
						rtc->reg[0x0D]|=0x08;
					else
						rtc->reg[0x0D]&=0x07;
					
					if(IsDlgButtonChecked(hDlg, IDC_RTC_E1))
						rtc->reg[0x0E]|=0x01;
					else
						rtc->reg[0x0E]&=0x0E;
					if(IsDlgButtonChecked(hDlg, IDC_RTC_E2))
						rtc->reg[0x0E]|=0x02;
					else
						rtc->reg[0x0E]&=0x0D;
					if(IsDlgButtonChecked(hDlg, IDC_RTC_E4))
						rtc->reg[0x0E]|=0x04;
					else
						rtc->reg[0x0E]&=0x0B;
					if(IsDlgButtonChecked(hDlg, IDC_RTC_E8))
						rtc->reg[0x0E]|=0x08;
					else
						rtc->reg[0x0E]&=0x07;
					
					if(IsDlgButtonChecked(hDlg, IDC_RTC_F1))
						rtc->reg[0x0F]|=0x01;
					else
						rtc->reg[0x0F]&=0x0E;
					if(IsDlgButtonChecked(hDlg, IDC_RTC_F2))
						rtc->reg[0x0F]|=0x02;
					else
						rtc->reg[0x0F]&=0x0D;
					if(IsDlgButtonChecked(hDlg, IDC_RTC_F4))
						rtc->reg[0x0F]|=0x04;
					else
						rtc->reg[0x0F]&=0x0B;
					if(IsDlgButtonChecked(hDlg, IDC_RTC_F8))
						rtc->reg[0x0F]|=0x08;
					else
						rtc->reg[0x0F]&=0x07;

					
					
					
				case IDCANCEL:
					EndDialog(hDlg, 1);
					return true;
				/*case IDC_MONTH:
					int dinmonth;
					dinmonth = S9xRTCDaysInMonth(
				*/
				default: return false;
			}
			case WM_HSCROLL:
				month=1+SendDlgItemMessage(hDlg,IDC_MONTH,TBM_GETPOS,0,0);
				SetDlgItemInt(hDlg,IDC_LBLMONTH,month,0);
				day=1+SendDlgItemMessage(hDlg,IDC_DAY,TBM_GETPOS,0,0);
				year=SendDlgItemMessage(hDlg,IDC_YEAR,TBM_GETPOS,0,0);
				hour=SendDlgItemMessage(hDlg,IDC_HOUR,TBM_GETPOS,0,0);
				minutes=SendDlgItemMessage(hDlg,IDC_MINUTE,TBM_GETPOS,0,0);
				seconds=SendDlgItemMessage(hDlg,IDC_SECOND,TBM_GETPOS,0,0);
				SetDlgItemInt(hDlg,IDC_LBLHOUR,hour,0);
				SetDlgItemInt(hDlg,IDC_LBLMINUTE,minutes,0);
				SetDlgItemInt(hDlg,IDC_LBLSECOND,seconds,0);
				if(year<5)
				{
					year+=95;
					SetDlgItemInt(hDlg,IDC_LBLYEAR,year+1900,0);
				}
				else
				{
					year-=5;
					SetDlgItemInt(hDlg,IDC_LBLYEAR,year+2000,0);
				}
				dayinmonth=S9xRTCDaysInMonth(month,year);
				if(day>dayinmonth)
				{
					day=dayinmonth;
					SendDlgItemMessage(hDlg,IDC_DAY, TBM_SETPOS,(WPARAM)(BOOL)TRUE,(LPARAM)day-1);
				}
				SendDlgItemMessage(hDlg,IDC_DAY, TBM_SETRANGE,(WPARAM)1,(LPARAM)MAKELONG(0,dayinmonth-1));
				SetDlgItemInt(hDlg,IDC_LBLDAY,day,0);
				return true;

		default: return false;
	}

}
#endif

int CALLBACK InfoProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
		case WM_INITDIALOG:
			{
				char temp[100];
				char romtext[4096];
				sprintf(romtext, "File: %s\r\nName: %s\r\n", Memory.ROMFilename, Memory.ROMName);
				sprintf(temp, "Speed: %02X/%s\r\nROM Map: %s\r\nType: %02x\r\n", Memory.ROMSpeed, ((Memory.ROMSpeed&0x10)!=0)?"FastROM":"SlowROM",((Memory.ROMSpeed&0x0F)!=0)?"HiROM":"LoROM",Memory.ROMType);
				strcat(romtext, temp);
				strcat(romtext, "Kart contents: ");
				strcat(romtext, Memory.KartContents ());
				strcat(romtext, "\r\nHeader ROM Size: ");
				strcat(romtext, Memory.Size());
				sprintf(temp, "\r\nCalculated ROM Size: %d Mbits", Memory.CalculatedSize/0x20000);
				strcat(romtext, temp);
				
				strcat(romtext, "\r\nSRAM size: ");
				strcat(romtext, Memory.StaticRAMSize ());
				strcat(romtext, "\r\nActual Checksum: ");
				sprintf(temp, "%04X", Memory.CalculatedChecksum);
				strcat(romtext, temp);
				strcat(romtext, "\r\nHeader Checksum: ");
				sprintf(temp, "%04X", Memory.ROMChecksum);
				strcat(romtext, temp);
				strcat(romtext, "\r\nHeader Checksum Compliment: ");
				sprintf(temp, "%04X", Memory.ROMComplementChecksum);
				strcat(romtext, temp);
				strcat(romtext, "\r\nOutput: ");
				if(Memory.ROMRegion>12||Memory.ROMRegion<2)
					strcat(romtext, "NTSC 60Hz");
				else strcat(romtext, "PAL 50Hz");
				strcat(romtext, "\r\nLicensee: ");

				int tmp=-1;
//				sscanf(Memory.CompanyId, "%02X", &tmp);
				if(Memory.CompanyId[0]=='0')
					tmp=0;
				if(Memory.CompanyId[0]=='1')
					tmp=16;
				if(Memory.CompanyId[0]=='2')
					tmp=32;
				if(Memory.CompanyId[0]=='3')
					tmp=48;
				if(Memory.CompanyId[0]=='4')
					tmp=64;
				if(Memory.CompanyId[0]=='5')
					tmp=80;
				if(Memory.CompanyId[0]=='6')
					tmp=96;
				if(Memory.CompanyId[0]=='7')
					tmp=112;
				if(Memory.CompanyId[0]=='8')
					tmp=128;
				if(Memory.CompanyId[0]=='9')
					tmp=144;
				if(Memory.CompanyId[0]=='A')
					tmp=160;
				if(Memory.CompanyId[0]=='B')
					tmp=176;
				if(Memory.CompanyId[0]=='C')
					tmp=192;
				if(Memory.CompanyId[0]=='D')
					tmp=208;
				if(Memory.CompanyId[0]=='E')
					tmp=224;
				if(Memory.CompanyId[0]=='F')
					tmp=240;
				if(tmp!=-1)
				{
					if(Memory.CompanyId[1]=='0')
						tmp+=0;
					else if(Memory.CompanyId[1]=='1')
						tmp+=1;
					else if(Memory.CompanyId[1]=='2')
						tmp+=2;
					else if(Memory.CompanyId[1]=='3')
						tmp+=3;
					else if(Memory.CompanyId[1]=='4')
						tmp+=4;
					else if(Memory.CompanyId[1]=='5')
						tmp+=5;
					else if(Memory.CompanyId[1]=='6')
						tmp+=6;
					else if(Memory.CompanyId[1]=='7')
						tmp+=7;
					else if(Memory.CompanyId[1]=='8')
						tmp+=8;
					else if(Memory.CompanyId[1]=='9')
						tmp+=9;
					else if(Memory.CompanyId[1]=='A')
						tmp+=10;
					else if(Memory.CompanyId[1]=='B')
						tmp+=11;
					else if(Memory.CompanyId[1]=='C')
						tmp+=12;
					else if(Memory.CompanyId[1]=='D')
						tmp+=13;
					else if(Memory.CompanyId[1]=='E')
						tmp+=14;
					else if(Memory.CompanyId[1]=='F')
						tmp+=15;
					else tmp=0;
				}
				else tmp=0;
				if(tmp==0)
					tmp=((Memory.ROMSpeed&0x0F)!=0)?Memory.ROM[0x0FFDA]:Memory.ROM[0x7FDA];
				switch(tmp)
//				switch(((Memory.ROMSpeed&0x0F)!=0)?Memory.ROM[0x0FFDA]:Memory.ROM[0x7FDA])
//				switch(atoi(Memory.CompanyId))
//				switch(((Memory.CompanyId[0]-'0')*16)+(Memory.CompanyId[1]-'0'))
				{
				case 0:strcat(romtext, "INVALID company");break;
				case 1:strcat(romtext, "Nintendo");break;
				case 2:strcat(romtext, NOTKNOWN "2");break;
				case 3:strcat(romtext, "Imagineer-Zoom");break;
				case 4:strcat(romtext, NOTKNOWN "4");break;
				case 5:strcat(romtext, "Zamuse");break;
				case 6:strcat(romtext, "Falcom");break;
				case 7:strcat(romtext, NOTKNOWN "7");break;
				case 8:strcat(romtext, "Capcom");break;
				case 9:strcat(romtext, "HOT-B");break;
				case 10:strcat(romtext, "Jaleco");break;
				case 11:strcat(romtext, "Coconuts");break;
				case 12:strcat(romtext, "Rage Software");break;
				case 13:strcat(romtext, NOTKNOWN "13");break;
				case 14:strcat(romtext, "Technos");break;
				case 15:strcat(romtext, "Mebio Software");break;
				case 16:strcat(romtext, NOTKNOWN "16");break;
				case 17:strcat(romtext, NOTKNOWN "17");break;
				case 18:strcat(romtext, "Gremlin Graphics");break;
				case 19:strcat(romtext, "Electronic Arts");break;
				case 20:strcat(romtext, NOTKNOWN "20");break;
				case 21:strcat(romtext, "COBRA Team");break;
				case 22:strcat(romtext, "Human/Field");break;
				case 23:strcat(romtext, "KOEI");break;
				case 24:strcat(romtext, "Hudson Soft");break;
				case 25:strcat(romtext, NOTKNOWN "25");break;
				case 26:strcat(romtext, "Yanoman");break;
				case 27:strcat(romtext, NOTKNOWN "27");break;
				case 28:strcat(romtext, "Tecmo");break;
				case 29:strcat(romtext, NOTKNOWN "29");break;
				case 30:strcat(romtext, "Open System");break;
				case 31:strcat(romtext, "Virgin Games");break;
				case 32:strcat(romtext, "KSS");break;
				case 33:strcat(romtext, "Sunsoft");break;
				case 34:strcat(romtext, "POW");break;
				case 35:strcat(romtext, "Micro World");break;
				case 36:strcat(romtext, NOTKNOWN "36");break;
				case 37:strcat(romtext, NOTKNOWN "37");break;
				case 38:strcat(romtext, "Enix");break;
				case 39:strcat(romtext, NOTKNOWN "39");break;
				case 40:strcat(romtext, "Kemco");break;
				case 41:strcat(romtext, "Seta Co.,Ltd.");break;
				case 42:strcat(romtext, NOTKNOWN "42");break;
				case 43:strcat(romtext, "Irem Japan");break;//Irem? Gun Force J
				case 44:strcat(romtext, NOTKNOWN "44");break;
				case 45:strcat(romtext, "Visit Co.,Ltd.");break;
				case 46:strcat(romtext, NOTKNOWN "46");break;
				case 47:strcat(romtext, NOTKNOWN "47");break;
				case 48:strcat(romtext, NOTKNOWN "48");break;
				case 49:strcat(romtext, "Carrozzeria");break;
				case 50:strcat(romtext, "Dynamic");break;
				case 51:strcat(romtext, "Nintendo");break;
				case 52:strcat(romtext, "Magifact");break;
				case 53:strcat(romtext, "Hect");break;
				case 54:strcat(romtext, NOTKNOWN "54");break;
				case 55:strcat(romtext, NOTKNOWN "55");break;
				case 56:strcat(romtext, "Capcom Europe");break;//Capcom? BOF2(E) MM7 (E)
				case 57:strcat(romtext, "Accolade Europe");break;//Accolade?Bubsy 2 (E)
				case 58:strcat(romtext, NOTKNOWN "58");break;
				case 59:strcat(romtext, NOTKNOWN "59");break;
				case 60:strcat(romtext, "Empire Software");break;
				case 61:strcat(romtext, "Loriciel");break;
				case 62:strcat(romtext, NOTKNOWN "62");break;
				case 63:strcat(romtext, NOTKNOWN "63");break;
				case 64:strcat(romtext, "Seika Corp.");break;
				case 65:strcat(romtext, "UBI Soft");break;
				case 66:strcat(romtext, NOTKNOWN "66");break;
				case 67:strcat(romtext, NOTKNOWN "67");break;
				case 68:strcat(romtext, "LifeFitness Exertainment");break;//?? Exertainment Mountain Bike Rally (U).zip
				case 69:strcat(romtext, NOTKNOWN "69");break;
				case 70:strcat(romtext, "System 3");break;
				case 71:strcat(romtext, "Spectrum Holobyte");break;
				case 72:strcat(romtext, NOTKNOWN "72");break;
				case 73:strcat(romtext, "Irem");break;
				case 74:strcat(romtext, NOTKNOWN "74");break;
				case 75:strcat(romtext, "Raya Systems/Sculptured Software");break;
				case 76:strcat(romtext, "Renovation Products");break;
				case 77:strcat(romtext, "Malibu Games/Black Pearl");break;
				case 78:strcat(romtext, NOTKNOWN "78");break;
				case 79:strcat(romtext, "U.S. Gold");break;
				case 80:strcat(romtext, "Absolute Entertainment");break;
				case 81:strcat(romtext, "Acclaim");break;
				case 82:strcat(romtext, "Activision");break;
				case 83:strcat(romtext, "American Sammy");break;
				case 84:strcat(romtext, "GameTek");break;
				case 85:strcat(romtext, "Hi Tech Expressions");break;
				case 86:strcat(romtext, "LJN Toys");break;
				case 87:strcat(romtext, NOTKNOWN "87");break;
				case 88:strcat(romtext, NOTKNOWN "88");break;
				case 89:strcat(romtext, NOTKNOWN "89");break;
				case 90:strcat(romtext, "Mindscape");break;
				case 91:strcat(romtext, NOTKNOWN "91");break;
				case 92:strcat(romtext, NOTKNOWN "92");break;
				case 93:strcat(romtext, "Tradewest");break;
				case 94:strcat(romtext, NOTKNOWN "94");break;
				case 95:strcat(romtext, "American Softworks Corp.");break;
				case 96:strcat(romtext, "Titus");break;
				case 97:strcat(romtext, "Virgin Interactive Entertainment");break;
				case 98:strcat(romtext, "Maxis");break;
				case 99:strcat(romtext, NOTKNOWN "99");break;
				case 100:strcat(romtext, NOTKNOWN "100");break;
				case 101:strcat(romtext, NOTKNOWN "101");break;
				case 102:strcat(romtext, NOTKNOWN "102");break;
				case 103:strcat(romtext, "Ocean");break;
				case 104:strcat(romtext, NOTKNOWN "104");break;
				case 105:strcat(romtext, "Electronic Arts");break;
				case 106:strcat(romtext, NOTKNOWN "106");break;
				case 107:strcat(romtext, "Laser Beam");break;
				case 108:strcat(romtext, NOTKNOWN "108");break;
				case 109:strcat(romtext, NOTKNOWN "109");break;
				case 110:strcat(romtext, "Elite");break;
				case 111:strcat(romtext, "Electro Brain");break;
				case 112:strcat(romtext, "Infogrames");break;
				case 113:strcat(romtext, "Interplay");break;
				case 114:strcat(romtext, "LucasArts");break;
				case 115:strcat(romtext, "Parker Brothers");break;
				case 116:strcat(romtext, NOTKNOWN "116");break;
				case 117:strcat(romtext, "STORM");break;
				case 118:strcat(romtext, NOTKNOWN "118");break;
				case 119:strcat(romtext, NOTKNOWN "119");break;
				case 120:strcat(romtext, "THQ Software");break;
				case 121:strcat(romtext, "Accolade Inc.");break;
				case 122:strcat(romtext, "Triffix Entertainment");break;
				case 123:strcat(romtext, NOTKNOWN "123");break;
				case 124:strcat(romtext, "Microprose");break;
				case 125:strcat(romtext, NOTKNOWN "125");break;
				case 126:strcat(romtext, NOTKNOWN "126");break;
				case 127:strcat(romtext, "Kemco");break;
				case 128:strcat(romtext, "Misawa");break;
				case 129:strcat(romtext, "Teichio");break;
				case 130:strcat(romtext, "Namco Ltd.");break;
				case 131:strcat(romtext, "Lozc");break;
				case 132:strcat(romtext, "Koei");break;
				case 133:strcat(romtext, NOTKNOWN "133");break;
				case 134:strcat(romtext, "Tokuma Shoten Intermedia");break;
				case 135:strcat(romtext, NOTKNOWN "135");break;
				case 136:strcat(romtext, "DATAM-Polystar");break;
				case 137:strcat(romtext, NOTKNOWN "137");break;
				case 138:strcat(romtext, NOTKNOWN "138");break;
				case 139:strcat(romtext, "Bullet-Proof Software");break;
				case 140:strcat(romtext, "Vic Tokai");break;
				case 141:strcat(romtext, NOTKNOWN "141");break;
				case 142:strcat(romtext, "Character Soft");break;
				case 143:strcat(romtext, "I\'\'Max");break;
				case 144:strcat(romtext, "Takara");break;
				case 145:strcat(romtext, "CHUN Soft");break;
				case 146:strcat(romtext, "Video System Co., Ltd.");break;
				case 147:strcat(romtext, "BEC");break;
				case 148:strcat(romtext, NOTKNOWN "148");break;
				case 149:strcat(romtext, "Varie");break;
				case 150:strcat(romtext, NOTKNOWN "150");break;
				case 151:strcat(romtext, "Kaneco");break;
				case 152:strcat(romtext, NOTKNOWN "152");break;
				case 153:strcat(romtext, "Pack in Video");break;
				case 154:strcat(romtext, "Nichibutsu");break;
				case 155:strcat(romtext, "TECMO");break;
				case 156:strcat(romtext, "Imagineer Co.");break;
				case 157:strcat(romtext, NOTKNOWN "157");break;
				case 158:strcat(romtext, NOTKNOWN "158");break;
				case 159:strcat(romtext, NOTKNOWN "159");break;
				case 160:strcat(romtext, "Telenet");break;
				case 161:strcat(romtext, NOTKNOWN "161");break;
				case 162:strcat(romtext, NOTKNOWN "162");break;
				case 163:strcat(romtext, NOTKNOWN "163");break;
				case 164:strcat(romtext, "Konami");break;
				case 165:strcat(romtext, "K.Amusement Leasing Co.");break;
				case 166:strcat(romtext, NOTKNOWN "166");break;
				case 167:strcat(romtext, "Takara");break;
				case 168:strcat(romtext, NOTKNOWN "168");break;
				case 169:strcat(romtext, "Technos Jap.");break;
				case 170:strcat(romtext, "JVC");break;
				case 171:strcat(romtext, NOTKNOWN "171");break;
				case 172:strcat(romtext, "Toei Animation");break;
				case 173:strcat(romtext, "Toho");break;
				case 174:strcat(romtext, NOTKNOWN "174");break;
				case 175:strcat(romtext, "Namco Ltd.");break;
				case 176:strcat(romtext, NOTKNOWN "176");break;
				case 177:strcat(romtext, "ASCII Co. Activison");break;
				case 178:strcat(romtext, "Bandai");break;
				case 179:strcat(romtext, NOTKNOWN "179");break;
				case 180:strcat(romtext, "Enix America");break;
				case 181:strcat(romtext, NOTKNOWN "181");break;
				case 182:strcat(romtext, "Halken");break;
				case 183:strcat(romtext, NOTKNOWN "183");break;
				case 184:strcat(romtext, NOTKNOWN "184");break;
				case 185:strcat(romtext, NOTKNOWN "185");break;
				case 186:strcat(romtext, "Culture Brain");break;
				case 187:strcat(romtext, "Sunsoft");break;
				case 188:strcat(romtext, "Toshiba EMI");break;
				case 189:strcat(romtext, "Sony Imagesoft");break;
				case 190:strcat(romtext, NOTKNOWN "190");break;
				case 191:strcat(romtext, "Sammy");break;
				case 192:strcat(romtext, "Taito");break;
				case 193:strcat(romtext, NOTKNOWN "193");break;
				case 194:strcat(romtext, "Kemco");break;
				case 195:strcat(romtext, "Square");break;
				case 196:strcat(romtext, "Tokuma Soft");break;
				case 197:strcat(romtext, "Data East");break;
				case 198:strcat(romtext, "Tonkin House");break;
				case 199:strcat(romtext, NOTKNOWN "199");break;
				case 200:strcat(romtext, "KOEI");break;
				case 201:strcat(romtext, NOTKNOWN "201");break;
				case 202:strcat(romtext, "Konami USA");break;
				case 203:strcat(romtext, "NTVIC");break;
				case 204:strcat(romtext, NOTKNOWN "204");break;
				case 205:strcat(romtext, "Meldac");break;
				case 206:strcat(romtext, "Pony Canyon");break;
				case 207:strcat(romtext, "Sotsu Agency/Sunrise");break;
				case 208:strcat(romtext, "Disco/Taito");break;
				case 209:strcat(romtext, "Sofel");break;
				case 210:strcat(romtext, "Quest Corp.");break;
				case 211:strcat(romtext, "Sigma");break;
				case 212:strcat(romtext, NOTKNOWN "212");break;
				case 213:strcat(romtext, NOTKNOWN "213");break;
				case 214:strcat(romtext, "Naxat");break;
				case 215:strcat(romtext, NOTKNOWN "215");break;
				case 216:strcat(romtext, "Capcom Co., Ltd.");break;
				case 217:strcat(romtext, "Banpresto");break;
				case 218:strcat(romtext, "Tomy");break;
				case 219:strcat(romtext, "Acclaim");break;
				case 220:strcat(romtext, NOTKNOWN "220");break;
				case 221:strcat(romtext, "NCS");break;
				case 222:strcat(romtext, "Human Entertainment");break;
				case 223:strcat(romtext, "Altron");break;
				case 224:strcat(romtext, "Jaleco");break;
				case 225:strcat(romtext, NOTKNOWN "225");break;
				case 226:strcat(romtext, "Yutaka");break;
				case 227:strcat(romtext, NOTKNOWN "227");break;
				case 228:strcat(romtext, "T&ESoft");break;
				case 229:strcat(romtext, "EPOCH Co.,Ltd.");break;
				case 230:strcat(romtext, NOTKNOWN "230");break;
				case 231:strcat(romtext, "Athena");break;
				case 232:strcat(romtext, "Asmik");break;
				case 233:strcat(romtext, "Natsume");break;
				case 234:strcat(romtext, "King Records");break;
				case 235:strcat(romtext, "Atlus");break;
				case 236:strcat(romtext, "Sony Music Entertainment");break;
				case 237:strcat(romtext, NOTKNOWN "237");break;
				case 238:strcat(romtext, "IGS");break;
				case 239:strcat(romtext, NOTKNOWN "239");break;
				case 240:strcat(romtext, NOTKNOWN "240");break;
				case 241:strcat(romtext, "Motown Software");break;
				case 242:strcat(romtext, "Left Field Entertainment");break;
				case 243:strcat(romtext, "Beam Software");break;
				case 244:strcat(romtext, "Tec Magik");break;
				case 245:strcat(romtext, NOTKNOWN "245");break;
				case 246:strcat(romtext, NOTKNOWN "246");break;
				case 247:strcat(romtext, NOTKNOWN "247");break;
				case 248:strcat(romtext, NOTKNOWN "248");break;
				case 249:strcat(romtext, "Cybersoft");break;
				case 250:strcat(romtext, NOTKNOWN "250");break;
				case 251:strcat(romtext, NOTKNOWN "251");break;
				case 252:strcat(romtext, NOTKNOWN "252");break;
				case 253:strcat(romtext, NOTKNOWN "253");break;
				case 254:strcat(romtext, NOTKNOWN "254");break;
				case 255:strcat(romtext, NOTKNOWN "255");break;
				default:strcat(romtext, NOTKNOWN);break;
				}

				strcat(romtext, "\r\nROM Version: ");
				sprintf(temp, "1.%d", ((Memory.ROMSpeed&0x0F)!=0)?Memory.ROM[0x0FFDB]:Memory.ROM[0x7FDB]);
				strcat(romtext, temp);
				strcat(romtext, "\r\nRegion: ");
				switch(Memory.ROMRegion)
				{
					case 0:
						strcat(romtext, "Japan");
						break;
					case 1:
						strcat(romtext, "USA/Canada");
						break;
					case 2:
						strcat(romtext, "Oceania, Europe, and Asia");
						break;
					case 3:
						strcat(romtext, "Sweden");
						break;
					case 4:
						strcat(romtext, "Finland");
						break;
					case 5:
						strcat(romtext, "Denmark");
						break;
					case 6:
						strcat(romtext, "France");
						break;
					case 7:
						strcat(romtext, "Holland");
						break;
					case 8:
						strcat(romtext, "Spain");
						break;
					case 9:
						strcat(romtext, "Germany, Austria, and Switzerland");
						break;
					case 10:
						strcat(romtext, "Italy");
						break;
					case 11:
						strcat(romtext, "Hong Kong and China");
						break;
					case 12:
						strcat(romtext, "Indonesia");
						break;
					case 13:
						strcat(romtext, "South Korea");
						break;
					case 14:strcat(romtext, "Unknown region 14");break;
					default:strcat(romtext, "Unknown region 15");break;
				}
				SendDlgItemMessage(hDlg, IDC_ROM_DATA, WM_SETTEXT, 0, (LPARAM)romtext);
				return true;
				break;
			}
		case WM_COMMAND:
			{
				switch(wParam)
				{
					case IDOK:
						EndDialog(hDlg, 0);
						return true;
						break;
					default: return false; break;
				}
			}
		default:return false;
	}
}